# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Sharpness Aware Minimization implementation."""

import copy

import tensorflow.compat.v2 as tf

from keras.src.engine import data_adapter
from keras.src.layers import deserialize as deserialize_layer
from keras.src.models import Model
from keras.src.saving.object_registration import register_keras_serializable
from keras.src.saving.serialization_lib import serialize_keras_object

# isort: off
from tensorflow.python.util.tf_export import keras_export


@register_keras_serializable()
@keras_export("keras.models.experimental.SharpnessAwareMinimization", v1=[])
class SharpnessAwareMinimization(Model):
    """Sharpness aware minimization (SAM) training flow.

    Sharpness-aware minimization (SAM) is a technique that improves the model
    generalization and provides robustness to label noise. Mini-batch splitting
    is proven to improve the SAM's performance, so users can control how mini
    batches are split via setting the `num_batch_splits` argument.

    Args:
      model: `tf.keras.Model` instance. The inner model that does the
        forward-backward pass.
      rho: float. The gradients scaling factor. Defaults to `0.05`.
      num_batch_splits: int. The number of mini batches to
        split into from each data batch. If None, batches are not split into
        sub-batches. Defaults to `None`.
      name: string. The name of the SAM model. Defaults to `None`.

    Reference:
      [Pierre Foret et al., 2020](https://arxiv.org/abs/2010.01412)
    """

    def __init__(self, model, rho=0.05, num_batch_splits=None, name=None):
        super().__init__(name=name)
        self.model = model
        self.rho = rho
        self.num_batch_splits = num_batch_splits

    def train_step(self, data):
        """The logic of one SAM training step.

        Args:
          data: A nested structure of `Tensor`s. It should be of structure
            (x, y, sample_weight) or (x, y).

        Returns:
          A dict mapping metric names to running average values.
        """
        x, y, sample_weight = data_adapter.unpack_x_y_sample_weight(data)

        if self.num_batch_splits is not None:
            x_split = tf.split(x, self.num_batch_splits)
            y_split = tf.split(y, self.num_batch_splits)
        else:
            x_split = [x]
            y_split = [y]

        gradients_all_batches = []
        pred_all_batches = []
        for x_batch, y_batch in zip(x_split, y_split):
            epsilon_w_cache = []
            with tf.GradientTape() as tape:
                pred = self.model(x_batch)
                loss = self.compiled_loss(y_batch, pred)
            pred_all_batches.append(pred)
            trainable_variables = self.model.trainable_variables
            gradients = tape.gradient(loss, trainable_variables)

            gradients_order2_norm = self._gradients_order2_norm(gradients)
            scale = self.rho / (gradients_order2_norm + 1e-12)

            for gradient, variable in zip(gradients, trainable_variables):
                epsilon_w = gradient * scale
                self._distributed_apply_epsilon_w(
                    variable, epsilon_w, tf.distribute.get_strategy()
                )
                epsilon_w_cache.append(epsilon_w)

            with tf.GradientTape() as tape:
                pred = self(x_batch)
                loss = self.compiled_loss(y_batch, pred)
            gradients = tape.gradient(loss, trainable_variables)
            if len(gradients_all_batches) == 0:
                for gradient in gradients:
                    gradients_all_batches.append([gradient])
            else:
                for gradient, gradient_all_batches in zip(
                    gradients, gradients_all_batches
                ):
                    gradient_all_batches.append(gradient)
            for variable, epsilon_w in zip(
                trainable_variables, epsilon_w_cache
            ):
                # Restore the variable to its original value before
                # `apply_gradients()`.
                self._distributed_apply_epsilon_w(
                    variable, -epsilon_w, tf.distribute.get_strategy()
                )

        gradients = []
        for gradient_all_batches in gradients_all_batches:
            gradients.append(tf.reduce_sum(gradient_all_batches, axis=0))
        self.optimizer.apply_gradients(zip(gradients, trainable_variables))

        pred = tf.concat(pred_all_batches, axis=0)
        self.compiled_metrics.update_state(y, pred, sample_weight)
        return {m.name: m.result() for m in self.metrics}

    def call(self, inputs):
        """Forward pass of SAM.

        SAM delegates the forward pass call to the wrapped model.

        Args:
          inputs: Tensor. The model inputs.

        Returns:
          A Tensor, the outputs of the wrapped model for given `inputs`.
        """
        return self.model(inputs)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "model": serialize_keras_object(self.model),
                "rho": self.rho,
            }
        )
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        # Avoid mutating the input dict.
        config = copy.deepcopy(config)
        model = deserialize_layer(
            config.pop("model"), custom_objects=custom_objects
        )
        config["model"] = model
        return super().from_config(config, custom_objects)

    def _distributed_apply_epsilon_w(self, var, epsilon_w, strategy):
        # Helper function to apply epsilon_w on model variables.
        if isinstance(
            tf.distribute.get_strategy(),
            (
                tf.distribute.experimental.ParameterServerStrategy,
                tf.distribute.experimental.CentralStorageStrategy,
            ),
        ):
            # Under PSS and CSS, the AggregatingVariable has to be kept in sync.
            def distribute_apply(strategy, var, epsilon_w):
                strategy.extended.update(
                    var,
                    lambda x, y: x.assign_add(y),
                    args=(epsilon_w,),
                    group=False,
                )

            tf.__internal__.distribute.interim.maybe_merge_call(
                distribute_apply, tf.distribute.get_strategy(), var, epsilon_w
            )
        else:
            var.assign_add(epsilon_w)

    def _gradients_order2_norm(self, gradients):
        norm = tf.norm(
            tf.stack([tf.norm(grad) for grad in gradients if grad is not None])
        )
        return norm




# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Benchmark for KPL implementation of weighted embedding column with
varying-length inputs."""

import tensorflow.compat.v2 as tf

import keras.src as keras
from keras.src.layers.preprocessing.benchmarks import (
    feature_column_benchmark as fc_bm,
)

# isort: off
from tensorflow.python.eager.def_function import (
    function as tf_function,
)

NUM_REPEATS = 10
BATCH_SIZES = [32, 256]


### KPL AND FC IMPLEMENTATION BENCHMARKS ###
def embedding_varlen(batch_size, max_length):
    """Benchmark a variable-length embedding."""
    # Data and constants.
    embedding_size = 32768
    data = fc_bm.create_data(
        max_length, batch_size * NUM_REPEATS, embedding_size - 1, dtype=int
    )
    weight = tf.ones_like(data, dtype=tf.float32)

    # Keras implementation
    data_input = keras.Input(
        shape=(None,), ragged=True, name="data", dtype=tf.int64
    )
    weight_input = keras.Input(
        shape=(None,), ragged=True, name="weight", dtype=tf.float32
    )
    embedded_data = keras.layers.Embedding(embedding_size, 256)(data_input)
    weighted_embedding = tf.multiply(
        embedded_data, tf.expand_dims(weight_input, -1)
    )
    reduced_embedding = tf.reduce_sum(weighted_embedding, axis=1)
    model = keras.Model([data_input, weight_input], reduced_embedding)

    # FC implementation
    fc = tf.feature_column.embedding_column(
        tf.feature_column.weighted_categorical_column(
            tf.feature_column.categorical_column_with_identity(
                "data", num_buckets=embedding_size - 1
            ),
            weight_feature_key="weight",
        ),
        dimension=256,
    )

    # Wrap the FC implementation in a tf.function for a fair comparison
    @tf_function()
    def fc_fn(tensors):
        fc.transform_feature(
            tf.__internal__.feature_column.FeatureTransformationCache(tensors),
            None,
        )

    # Benchmark runs
    keras_data = {"data": data, "weight": weight}
    k_avg_time = fc_bm.run_keras(keras_data, model, batch_size, NUM_REPEATS)

    fc_data = {"data": data.to_sparse(), "weight": weight.to_sparse()}
    fc_avg_time = fc_bm.run_fc(fc_data, fc_fn, batch_size, NUM_REPEATS)

    return k_avg_time, fc_avg_time


class BenchmarkLayer(fc_bm.LayerBenchmark):
    """Benchmark the layer forward pass."""

    def benchmark_layer(self):
        for batch in BATCH_SIZES:
            name = f"weighted_embedding|varlen|batch_{batch}"
            k_time, f_time = embedding_varlen(batch_size=batch, max_length=256)
            self.report(name, k_time, f_time, NUM_REPEATS)


if __name__ == "__main__":
    tf.test.main()




# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras SavedModel serialization.

TODO (kathywu): Move to layer_serialization.py. Some model-specific logic should
go to model_serialization.py.
"""

import functools
import threading
import weakref

import tensorflow.compat.v1.logging as logging
import tensorflow.compat.v2 as tf

from keras.src import backend
from keras.src.engine import base_layer_utils
from keras.src.engine import input_spec
from keras.src.mixed_precision import autocast_variable
from keras.src.saving.legacy import saving_utils
from keras.src.saving.legacy.saved_model import constants
from keras.src.saving.legacy.saved_model import load as keras_load
from keras.src.saving.legacy.saved_model import serialized_attributes
from keras.src.saving.legacy.saved_model import utils
from keras.src.utils import layer_utils
from keras.src.utils import tf_contextlib
from keras.src.utils import tf_utils
from keras.src.utils import version_utils
from keras.src.utils.generic_utils import LazyLoader

# To avoid circular dependencies between keras/engine and keras/saving,
# code in keras/saving must delay imports.

# TODO(b/134426265): Switch back to single-quotes to match the rest of the file
# once the issue with copybara is fixed.

base_layer = LazyLoader("base_layer", globals(), "keras.src.engine.base_layer")
metrics = LazyLoader("metrics", globals(), "keras.src.metrics")
input_layer = LazyLoader("input_layer", globals(), "keras.src.engine.input_layer")
training_lib = LazyLoader("training_lib", globals(), "keras.src.engine.training")
sequential_lib = LazyLoader(
    "sequential_lib", globals(), "keras.src.engine.sequential"
)


def should_skip_serialization(layer):
    """Skip serializing extra objects and functions if layer inputs aren't
    set."""
    saved_model_input_spec_set = (
        isinstance(layer, training_lib.Model)
        and layer._saved_model_inputs_spec is not None
    )
    if not layer.built and not saved_model_input_spec_set:
        logging.warning(
            "Skipping full serialization of Keras layer {}, because "
            "it is not built.".format(layer)
        )
        return True
    return False


def _filter_shards(variables):
    return [var for var in variables if not hasattr(var, "_sharded_container")]


def wrap_layer_objects(layer, serialization_cache):
    """Returns extra trackable objects to attach to the serialized layer.

    Args:
      layer: Keras Layer object.
      serialization_cache: Dictionary shared between all objects during
        serialization.

    Returns:
      A dictionary containing all checkpointable objects from a
      SerializedAttributes object. See LayerAttributes and ModelAttributes for
      entire list of objects
    """
    # Wrap all regularization losses as tf.functions.
    # First, generate list of all regularization losses in this layer and
    # sublayers.
    all_losses = layer._callable_losses[:]
    for child_layer in utils.list_all_layers(layer):
        all_losses.extend(child_layer._callable_losses)
    # Next, wrap all loss functions as tf.functions. Use the serialization cache
    # to store already-wrapped functions.
    keras_loss_cache = serialization_cache.setdefault("keras_losses", {})
    wrapped_loss_functions = []
    for loss_fn in all_losses:
        if loss_fn in keras_loss_cache:
            wrapped_loss_functions.append(keras_loss_cache[loss_fn])
        else:
            wrapped_loss = _wrap_unconditional_loss(
                loss_fn, len(keras_loss_cache)
            )
            keras_loss_cache[loss_fn] = wrapped_loss
            wrapped_loss_functions.append(wrapped_loss)
    wrapped_layer_losses = [
        keras_loss_cache[fn] for fn in layer._callable_losses[:]
    ]

    layer_metrics = tf.__internal__.tracking.wrap(
        {m.name: m for m in layer._metrics}
    )

    # Avoid duplicate creation of shard Variables on loading.
    # `layer.variables` will return the shard Variables rather than the
    # ShardedVariables (b/224541446), but Keras loading will create new
    # ShardedVariables (and thus shard Variables) from Keras metadata if needed.
    # There's no need to also save the shard Variables here, so filter them out.
    variables = _filter_shards(layer.variables)
    trainable_variables = _filter_shards(layer.trainable_variables)
    non_trainable_variables = _filter_shards(layer.non_trainable_variables)
    return dict(
        variables=tf.__internal__.tracking.wrap(variables),
        trainable_variables=tf.__internal__.tracking.wrap(trainable_variables),
        non_trainable_variables=tf.__internal__.tracking.wrap(
            non_trainable_variables
        ),
        layers=tf.__internal__.tracking.wrap(utils.list_all_layers(layer)),
        metrics=tf.__internal__.tracking.wrap(layer.metrics),
        regularization_losses=tf.__internal__.tracking.wrap(
            wrapped_loss_functions
        ),
        layer_regularization_losses=tf.__internal__.tracking.wrap(
            wrapped_layer_losses
        ),
        layer_metrics=layer_metrics,
    )


def wrap_layer_functions(layer, serialization_cache):
    """Returns dict of wrapped layer call function and losses in tf.functions.

    Args:
      layer: Keras Layer object.
      serialization_cache: Dictionary shared between all objects during
        serialization.

    Returns:
      A dictionary containing all keras tf.functions to serialize. See
      LayerAttributes and ModelAttributes for the list of all attributes.
    """
    # Since Sequential models may be modified in place using model.add() or
    # model.pop(), don't use saved functions.
    if isinstance(layer, keras_load.RevivedLayer) and not isinstance(
        layer, sequential_lib.Sequential
    ):
        return {
            fn_name: getattr(layer.keras_api, fn_name, None)
            for fn_name in serialized_attributes.LayerAttributes.all_functions
        }

    # Reset the losses of the layer and its children. The call function in each
    # child layer is replaced with tf.functions.
    original_fns = _replace_child_layer_functions(layer, serialization_cache)
    original_losses = _reset_layer_losses(layer)

    # Wrap all the layer call and activity regularizer functions.

    # Use LayerCallCollection to ensure that all layer call functions (__call__,
    # call with losses) are traced with the same inputs.
    call_collection = LayerCallCollection(layer)
    call_fn_with_losses = call_collection.add_function(
        _wrap_call_and_conditional_losses(layer),
        f"{layer.name}_layer_call_and_return_conditional_losses",
        # If any of this layer's child layers use the training arg, the traced
        # call functions of this layer will have a training keyword argument. If
        # the original layer does not expect the training arg, then it will have
        # to be removed (by setting `match_layer_training_arg`).
        match_layer_training_arg=True,
    )
    call_fn = call_collection.add_function(
        _extract_outputs_from_fn(layer, call_fn_with_losses),
        f"{layer.name}_layer_call_fn",
        # Since `call_fn` wraps call_fn_with_losses and not the original call
        # function, `match_layer_training_arg` should be set to False.
        match_layer_training_arg=False,
    )

    fns = {
        "call_and_return_conditional_losses": call_fn_with_losses,
        "__call__": call_fn,
    }

    if layer._activity_regularizer is not None:
        fns["activity_regularizer_fn"] = _wrap_activity_regularizer(layer)
        fns[
            "call_and_return_all_conditional_losses"
        ] = call_collection.add_function(
            _append_activity_regularizer_loss(
                layer, call_fn_with_losses, fns["activity_regularizer_fn"]
            ),
            f"{layer.name}_layer_call_and_return_all_conditional_losses",
            match_layer_training_arg=False,
        )
    else:
        fns["activity_regularizer_fn"] = None
        fns["call_and_return_all_conditional_losses"] = call_fn_with_losses

    # Manually trigger traces before restoring the overwritten functions. The
    # functions are traced within the layer call context to ensure that layer
    # functions (e.g. add_loss) behave as though running in graph mode.
    with tracing_scope():
        call_collection.trace_with_input_signature()
        with base_layer_utils.call_context().enter(
            layer, inputs=None, build_graph=True, training=None, saving=True
        ):
            for fn in fns.values():
                if fn is not None and not isinstance(fn, LayerCall):
                    fn.get_concrete_function()

    # Restore overwritten functions and losses
    _restore_child_layer_functions(original_fns)
    _restore_layer_losses(original_losses)

    return fns


def default_save_signature(layer):
    original_losses = _reset_layer_losses(layer)
    fn = saving_utils.trace_model_call(layer)
    _restore_layer_losses(original_losses)
    return fn


def _replace_child_layer_functions(layer, serialization_cache):
    """Replaces functions in the children layers with wrapped tf.functions.

    This step allows functions from parent layers to reference the wrapped
    functions from their children layers instead of retracing the ops.

    This function also resets all losses stored in the layer. These are stored
    in the returned dictionary. Use `_restore_child_layer_functions` to restore
    the original attributes.

    Args:
      layer: Keras Layer object.
      serialization_cache: Dictionary shared between all objects during
        serialization.

    Returns:
      Dictionary mapping layer objects -> original functions and losses:
        { Child layer 1: {
            'losses': Original losses,
            'call': Original call function
            '_activity_regularizer': Original activity regularizer},
          Child layer 2: ...
        }
    """

    original_fns = {}

    def replace_layer_functions(child_layer, serialized_fns):
        """Replaces layer call and activity regularizer with wrapped
        functions."""
        original_fns[child_layer] = {
            "call": child_layer.call,
            "_activity_regularizer": child_layer._activity_regularizer,
        }
        with utils.no_automatic_dependency_tracking_scope(child_layer):
            try:
                child_layer._activity_regularizer = serialized_fns.get(
                    "activity_regularizer_fn"
                )
            except AttributeError:
                # Some layers have an unsettable activity regularizer.
                pass
            child_layer.call = utils.use_wrapped_call(
                child_layer,
                serialized_fns["call_and_return_conditional_losses"],
                child_layer._call_spec,
                default_training_value=False,
            )

    def replace_metric_functions(child_layer, serialized_fns):
        """Replaces metric functions with wrapped functions."""
        original_fns[child_layer] = {
            "__call__": child_layer.__call__,
            "result": child_layer.result,
            "update_state": child_layer.update_state,
        }
        with utils.no_automatic_dependency_tracking_scope(child_layer):
            child_layer.__call__ = serialized_fns["__call__"]
            child_layer.result = serialized_fns["result"]
            child_layer.update_state = serialized_fns["update_state"]

    for child_layer in utils.list_all_layers(layer):
        if isinstance(child_layer, input_layer.InputLayer):
            continue

        if child_layer not in serialization_cache[constants.KERAS_CACHE_KEY]:
            serialized_functions = child_layer._trackable_saved_model_saver._get_serialized_attributes(  # noqa: E501
                serialization_cache
            ).functions
        else:
            serialized_functions = serialization_cache[
                constants.KERAS_CACHE_KEY
            ][child_layer].functions
        if not serialized_functions:
            # This indicates either:
            #   - circular dependency, which means the current layer's functions
            #     should be wrapped first.
            #   - Child layer's inputs are not defined, so its functions have
            #     not been wrapped. In this case, no replacement is necessary so
            #     move on to the next child.
            continue

        if isinstance(child_layer, metrics.Metric):
            replace_metric_functions(child_layer, serialized_functions)
        else:
            replace_layer_functions(child_layer, serialized_functions)

    return original_fns


def _restore_child_layer_functions(original_fns):
    """Restores attributes replaced with `_replace_child_layer_functions`."""
    for child_layer, fns in original_fns.items():
        with utils.no_automatic_dependency_tracking_scope(child_layer):
            for fn_name, fn in fns.items():
                try:
                    setattr(child_layer, fn_name, fn)
                except AttributeError:
                    # In the case of _activity_regularizer, setting the
                    # attribute may be disallowed.
                    pass


def _reset_layer_losses(parent_layer):
    """Resets losses of layer and its sublayers, and returns original losses."""
    losses_dict = {}
    for layer in utils.list_all_layers_and_sublayers(parent_layer):
        losses_dict[layer] = {
            "losses": layer._losses[:],
            "eager_losses": layer._eager_losses[:],
        }
        with utils.no_automatic_dependency_tracking_scope(layer):
            layer._losses = []
            layer._eager_losses = []
    return losses_dict


def _restore_layer_losses(losses_dict):
    for layer in losses_dict:
        with utils.no_automatic_dependency_tracking_scope(layer):
            layer._losses = losses_dict[layer]["losses"]
            layer._eager_losses = losses_dict[layer]["eager_losses"]


class LayerTracingContext(threading.local):
    def __init__(self):
        super().__init__()
        self.enable_call_tracing = False
        self.trace_queue = []


_thread_local_data = LayerTracingContext()


@tf_contextlib.contextmanager
def tracing_scope():
    """Enables tracing scope."""
    # This enables the LayerCallCollection's tracing mechanism to trace all call
    # functions in the collection.
    previous_value = _thread_local_data.enable_call_tracing
    previous_queue = _thread_local_data.trace_queue
    try:
        _thread_local_data.enable_call_tracing = True
        _thread_local_data.trace_queue = []
        yield
    finally:
        # Run traces from the queue.
        while _thread_local_data.trace_queue:
            fn, args, kwargs, training = _thread_local_data.trace_queue.pop(0)
            if training is not None:
                with backend.deprecated_internal_learning_phase_scope(training):
                    fn.get_concrete_function(*args, **kwargs)
            else:
                fn.get_concrete_function(*args, **kwargs)
        _thread_local_data.trace_queue = previous_queue
        _thread_local_data.enable_call_tracing = previous_value


def add_trace_to_queue(fn, args, kwargs, training=None):
    if tracing_enabled():
        _thread_local_data.trace_queue.append(
            (fn, args[:], kwargs.copy(), training)
        )


def tracing_enabled():
    """Whether to add extra traces to the queue."""
    return _thread_local_data.enable_call_tracing


class LayerCallCollection:
    """Groups wrapped layer call functions.

    This is used to ensure that all layer call functions are traced with the
    same inputs-
      - call
      - call_and_return_conditional_losses
      - call_and_return_all_conditional_losses
    """

    def __init__(self, layer):
        self.layer = layer

        self.layer_call_method = _get_layer_call_method(layer)
        self._expects_training_arg = utils.layer_uses_training_bool(layer)
        self._call_spec = layer._call_spec

        # Create new call spec if the layer itself does not accept a training
        # arg, but one of its child layers does. When this layer's call
        # functions are traced, they will be traced with an added `training`
        # keyword argument.
        if not self.layer._expects_training_arg and self._expects_training_arg:
            arg_spec = utils.set_training_arg_spec(
                self._call_spec.full_argspec, False
            )
            self._call_spec = layer_utils.CallFunctionSpec(arg_spec)

        self._layer_inputs = self._get_layer_inputs(layer)
        self._functions = weakref.WeakValueDictionary()

        # Get the input argument name from the args.
        if self._call_spec.arg_names:
            self._input_arg_name = self._call_spec.arg_names[0]
        else:
            # Layer could be defined with only varargs, in which case use a
            # default name.
            self._input_arg_name = "inputs"

    def _get_layer_inputs(self, layer):
        """Inspects layer object and returns the inferred input signature.

        Args:
          layer: Layer object.

        Returns:
          List of possibly nested TensorSpecs of the layer call function inputs
          in the form of `(args, kwargs)`
        """
        if (
            isinstance(layer.call, tf.__internal__.function.Function)
            and layer.call.input_signature is not None
        ):
            return layer.call.input_signature, {}
        elif isinstance(layer, training_lib.Model):
            return saving_utils.model_call_inputs(layer)
        elif (
            layer.input_spec is not None
            and layer._use_input_spec_as_call_signature
        ):

            def to_tensor_spec_or_none(x):
                spec = input_spec.to_tensor_spec(x, layer._compute_dtype)
                # If the shape is too general (e.g. multiple dimensions are
                # allowed), return None so that separate functions can be
                # generated for each inferred input signature.
                # TODO(b/134962016): currently partial signatures are not
                # supported.
                if spec.shape == tf.TensorShape(None):
                    return None, None
                return spec

            input_signature = [
                tf.nest.map_structure(to_tensor_spec_or_none, layer.input_spec)
            ]

            return input_signature, {}
        else:
            return None, None

    def add_trace(self, *args, **kwargs):
        """Traces all functions with the same args and kwargs.

        Args:
          *args: Positional args passed to the original function.
          **kwargs: Keyword args passed to the original function.
        """
        args = list(args)
        kwargs = kwargs.copy()

        for fn in self._functions.values():
            # TODO(kathywu): Replace arguments with broader shapes defined in
            # the input signature.
            if self._expects_training_arg:

                def trace_with_training(value, fn=fn):
                    nonlocal args, kwargs
                    (args, kwargs,) = self._call_spec.set_arg_value(
                        "training", value, args, kwargs, inputs_in_args=True
                    )
                    add_trace_to_queue(fn, args, kwargs, value)

                trace_with_training(True)
                trace_with_training(False)
            else:
                add_trace_to_queue(fn, args, kwargs)

    def training_arg_was_passed(self, args, kwargs):
        return self._call_spec.arg_was_passed(
            "training", args, kwargs, inputs_in_args=True
        )

    def get_training_arg_value(self, args, kwargs):
        try:
            return self._call_spec.get_arg_value(
                "training", args, kwargs, inputs_in_args=True
            )
        except KeyError:  # Training is not in args or kwargs.
            return None

    def get_input_arg_value(self, args, kwargs):
        return self._call_spec.get_arg_value(
            self._input_arg_name, args, kwargs, inputs_in_args=True
        )

    def _maybe_wrap_with_training_arg(self, call_fn, match_layer_training_arg):
        """Wraps call function with added training argument if necessary."""
        if not self.layer._expects_training_arg and self._expects_training_arg:
            # Add training arg to wrapper function.
            def wrap_with_training_arg(*args, **kwargs):
                if match_layer_training_arg:
                    # Remove the training value, since the original call_fn does
                    # not expect a training arg. Instead, the training value
                    # will be propagated using the call context created in
                    # LayerCall.
                    args = list(args)
                    kwargs = kwargs.copy()
                    (args, kwargs,) = self._call_spec.set_arg_value(
                        "training",
                        None,
                        args,
                        kwargs,
                        inputs_in_args=True,
                        pop_kwarg_if_none=True,
                    )
                return call_fn(*args, **kwargs)

            return tf.__internal__.decorator.make_decorator(
                target=call_fn,
                decorator_func=wrap_with_training_arg,
                decorator_argspec=self._call_spec.full_argspec,
            )

        return call_fn

    def add_function(self, call_fn, name, match_layer_training_arg):
        """Adds a layer call function to the collection.

        Args:
          call_fn: a python function
          name: Name of call function
          match_layer_training_arg: If True, removes the `training` from the
            function arguments when calling `call_fn`.

        Returns:
          LayerCall (tf.function)
        """
        fn = LayerCall(
            self,
            self._maybe_wrap_with_training_arg(
                call_fn, match_layer_training_arg
            ),
            name,
        )
        self._functions[name] = fn.wrapped_call
        return fn

    def trace_with_input_signature(self):
        """Trace with the layer/models inferred input signature if possible."""
        if self._layer_inputs[0] is None:
            return

        args, kwargs = self._layer_inputs
        if self._expects_training_arg:
            args, kwargs = self._call_spec.set_arg_value(
                "training", False, args, kwargs, inputs_in_args=True
            )
        if None not in tf.nest.flatten([args, kwargs]):
            # Manually add traces for layers that have keyword arguments and
            # have a fully defined input signature.
            self.add_trace(*args, **kwargs)


def _filtered_inputs(inputs):
    return list(filter(tf_utils.is_tensor_or_variable, tf.nest.flatten(inputs)))


def layer_call_wrapper(call_collection, method, name):
    """Ensures layer losses are kept the same, and runs method in call
    context."""

    # Create wrapper that deals with losses and call context.
    def wrapper(*args, **kwargs):
        """Calls method within call context."""
        layer = call_collection.layer
        training = None
        inputs = _filtered_inputs([args, kwargs])

        if (args or kwargs) and call_collection.training_arg_was_passed(
            args, kwargs
        ):
            training = call_collection.get_training_arg_value(args, kwargs)

        original_losses = _reset_layer_losses(layer)
        with base_layer_utils.call_context().enter(
            layer,
            inputs=inputs,
            build_graph=False,
            training=training,
            saving=True,
        ):
            with autocast_variable.enable_auto_cast_variables(
                layer._compute_dtype_object
            ):
                ret = method(*args, **kwargs)
        _restore_layer_losses(original_losses)
        return ret

    # Rename to `name`, since tf.function doesn't have a name argument. Without
    # this, all functions returned by this method will be named "call", which
    # would be a nightmare to debug.
    fn = tf.__internal__.decorator.make_decorator(
        target=method, decorator_func=wrapper
    )
    fn.__name__ = name
    return fn


class LayerCall:
    """Function that triggers traces of other functions in the same
    collection."""

    def __init__(self, call_collection, call_fn, name):
        """Initializes a LayerCall object.

        Args:
          call_collection: a LayerCallCollection, which contains the other layer
            call functions (e.g. call_with_conditional_losses, call). These
            functions should be traced with the same arguments.
          call_fn: A call function.
          name: Name of the call function.
        """
        self.call_collection = call_collection
        self.wrapped_call = tf.function(
            layer_call_wrapper(call_collection, call_fn, name)
        )

    def _maybe_trace(self, args, kwargs):
        # Trigger traces of other call functions + extra training-arg traces.
        if tracing_enabled():
            self.call_collection.add_trace(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        self._maybe_trace(args, kwargs)
        return self.wrapped_call(*args, **kwargs)

    def get_concrete_function(self, *args, **kwargs):
        self._maybe_trace(args, kwargs)
        return self.wrapped_call.get_concrete_function(*args, **kwargs)


def _wrap_call_and_conditional_losses(layer):
    """Wraps call function that returns a tuple of (outputs, losses).

    The losses returned are conditional on the inputs passed to the call
    function.  Unconditional losses (e.g. weight regularizeration) are wrapped
    separately.

    Args:
      layer: a Keras layer object

    Returns:
      python call function that returns outputs and conditional losses --
      excludes activity regularizer
    """
    # Create function that generates both outputs and losses
    layer_call = _get_layer_call_method(layer)

    def call_and_return_conditional_losses(*args, **kwargs):
        """Returns layer (call_output, conditional losses) tuple."""
        call_output = layer_call(*args, **kwargs)
        if version_utils.is_v1_layer_or_model(layer):
            conditional_losses = layer.get_losses_for(
                _filtered_inputs([args, kwargs])
            )
        else:
            conditional_losses = [
                l for l in layer.losses if not hasattr(l, "_unconditional_loss")
            ]
        return call_output, conditional_losses

    return _create_call_fn_decorator(layer, call_and_return_conditional_losses)


def _extract_outputs_from_fn(layer, call_and_return_conditional_losses):
    """Returns a function that returns only call function outputs."""
    if isinstance(layer, keras_load.RevivedLayer):
        return layer.keras_api.__call__

    def call(inputs, *args, **kwargs):
        return call_and_return_conditional_losses(inputs, *args, **kwargs)[0]

    return _create_call_fn_decorator(layer, call)


def _append_activity_regularizer_loss(
    layer, call_fn_with_losses, activity_regularizer_fn
):
    """Appends activity regularizer loss to losses returned by the wrapped
    fn."""

    def fn(inputs, *args, **kwargs):
        outputs, losses = call_fn_with_losses(inputs, *args, **kwargs)
        losses.append(activity_regularizer_fn(outputs))
        return outputs, losses

    return _create_call_fn_decorator(layer, fn)


def _create_call_fn_decorator(layer, wrapped_call):
    call_fn = _get_layer_call_method(layer)
    fn, arg_spec = utils.maybe_add_training_arg(
        layer._call_spec,
        wrapped_call,
        layer._expects_training_arg,
        default_training_value=False,
    )
    return tf.__internal__.decorator.make_decorator(
        target=call_fn, decorator_func=fn, decorator_argspec=arg_spec
    )


def _wrap_unconditional_loss(loss_fn, index):
    """Wraps callable/unconditional loss, returning a serializable function."""
    # Extract original loss function from partial function
    fn = loss_fn.args[0] if isinstance(loss_fn, functools.partial) else loss_fn
    if isinstance(fn, tf.__internal__.function.Function):
        return fn
    else:
        return tf.__internal__.function.Function(
            fn, f"loss_fn_{index}", input_signature=[]
        )


def _wrap_activity_regularizer(layer):
    """Wraps the activity regularizer."""

    if isinstance(
        layer._activity_regularizer, tf.__internal__.function.Function
    ):
        return layer._activity_regularizer
    return tf.__internal__.function.Function(
        layer._activity_regularizer,
        f"{layer.name}_activity_regularizer",
        input_signature=[
            tf.TensorSpec(None, layer._compute_dtype or backend.floatx())
        ],
    )


def _get_layer_call_method(layer):
    if isinstance(layer.call, (tf.__internal__.function.Function)):
        return layer.call.python_function
    return layer.call




# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras hashed crossing preprocessing layer."""


import tensorflow.compat.v2 as tf

from keras.src import backend
from keras.src.engine import base_layer
from keras.src.engine import base_preprocessing_layer
from keras.src.layers.preprocessing import preprocessing_utils as utils
from keras.src.utils import layer_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export

INT = utils.INT
ONE_HOT = utils.ONE_HOT


@keras_export(
    "keras.layers.HashedCrossing",
    "keras.layers.experimental.preprocessing.HashedCrossing",
    v1=[],
)
class HashedCrossing(base_layer.Layer):
    """A preprocessing layer which crosses features using the "hashing trick".

    This layer performs crosses of categorical features using the "hashing
    trick". Conceptually, the transformation can be thought of as:
    `hash(concatenate(features)) % num_bins`.

    This layer currently only performs crosses of scalar inputs and batches of
    scalar inputs. Valid input shapes are `(batch_size, 1)`, `(batch_size,)` and
    `()`.

    For an overview and full list of preprocessing layers, see the preprocessing
    [guide](https://www.tensorflow.org/guide/keras/preprocessing_layers).

    Args:
        num_bins: Number of hash bins.
        output_mode: Specification for the output of the layer. Values can be
            `"int"`, or `"one_hot"` configuring the layer as follows:
            - `"int"`: Return the integer bin indices directly.
            - `"one_hot"`: Encodes each individual element in the input into an
                array the same size as `num_bins`, containing a 1 at the input's
                bin index. Defaults to `"int"`.
        sparse: Boolean. Only applicable to `"one_hot"` mode. If `True`,
            returns a `SparseTensor` instead of a dense `Tensor`.
            Defaults to `False`.
        **kwargs: Keyword arguments to construct a layer.

    Examples:

    **Crossing two scalar features.**

    >>> layer = tf.keras.layers.HashedCrossing(
    ...     num_bins=5)
    >>> feat1 = tf.constant(['A', 'B', 'A', 'B', 'A'])
    >>> feat2 = tf.constant([101, 101, 101, 102, 102])
    >>> layer((feat1, feat2))
    <tf.Tensor: shape=(5,), dtype=int64, numpy=array([1, 4, 1, 1, 3])>

    **Crossing and one-hotting two scalar features.**

    >>> layer = tf.keras.layers.HashedCrossing(
    ...     num_bins=5, output_mode='one_hot')
    >>> feat1 = tf.constant(['A', 'B', 'A', 'B', 'A'])
    >>> feat2 = tf.constant([101, 101, 101, 102, 102])
    >>> layer((feat1, feat2))
    <tf.Tensor: shape=(5, 5), dtype=float32, numpy=
      array([[0., 1., 0., 0., 0.],
             [0., 0., 0., 0., 1.],
             [0., 1., 0., 0., 0.],
             [0., 1., 0., 0., 0.],
             [0., 0., 0., 1., 0.]], dtype=float32)>
    """

    def __init__(self, num_bins, output_mode="int", sparse=False, **kwargs):
        # By default, output int64 when output_mode="int" and floats otherwise.
        if "dtype" not in kwargs or kwargs["dtype"] is None:
            kwargs["dtype"] = (
                tf.int64 if output_mode == INT else backend.floatx()
            )

        super().__init__(**kwargs)
        base_preprocessing_layer.keras_kpl_gauge.get_cell("HashedCrossing").set(
            True
        )

        # Check dtype only after base layer parses it; dtype parsing is complex.
        if (
            output_mode == INT
            and not tf.as_dtype(self.compute_dtype).is_integer
        ):
            input_dtype = kwargs["dtype"]
            raise ValueError(
                "When `output_mode='int'`, `dtype` should be an integer "
                f"type. Received: dtype={input_dtype}"
            )

        # "output_mode" must be one of (INT, ONE_HOT)
        layer_utils.validate_string_arg(
            output_mode,
            allowable_strings=(INT, ONE_HOT),
            layer_name=self.__class__.__name__,
            arg_name="output_mode",
        )

        self.num_bins = num_bins
        self.output_mode = output_mode
        self.sparse = sparse

    def call(self, inputs):
        # Convert all inputs to tensors and check shape. This layer only
        # supports sclars and batches of scalars for the initial version.
        self._check_at_least_two_inputs(inputs)
        inputs = [utils.ensure_tensor(x) for x in inputs]
        self._check_input_shape_and_type(inputs)

        # Uprank to rank 2 for the cross_hashed op.
        rank = inputs[0].shape.rank
        if rank < 2:
            inputs = [utils.expand_dims(x, -1) for x in inputs]
        if rank < 1:
            inputs = [utils.expand_dims(x, -1) for x in inputs]

        # Perform the cross and convert to dense
        outputs = tf.sparse.cross_hashed(inputs, self.num_bins)
        outputs = tf.sparse.to_dense(outputs)

        # Fix output shape and downrank to match input rank.
        if rank == 2:
            # tf.sparse.cross_hashed output shape will always be None on the
            # last dimension. Given our input shape restrictions, we want to
            # force shape 1 instead.
            outputs = tf.reshape(outputs, [-1, 1])
        elif rank == 1:
            outputs = tf.reshape(outputs, [-1])
        elif rank == 0:
            outputs = tf.reshape(outputs, [])

        # Encode outputs.
        return utils.encode_categorical_inputs(
            outputs,
            output_mode=self.output_mode,
            depth=self.num_bins,
            sparse=self.sparse,
            dtype=self.compute_dtype,
        )

    def compute_output_shape(self, input_shapes):
        self._check_at_least_two_inputs(input_shapes)
        return utils.compute_shape_for_encode_categorical(input_shapes[0])

    def compute_output_signature(self, input_specs):
        input_shapes = [x.shape.as_list() for x in input_specs]
        output_shape = self.compute_output_shape(input_shapes)
        if self.sparse or any(
            isinstance(x, tf.SparseTensorSpec) for x in input_specs
        ):
            return tf.SparseTensorSpec(
                shape=output_shape, dtype=self.compute_dtype
            )
        return tf.TensorSpec(shape=output_shape, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_bins": self.num_bins,
                "output_mode": self.output_mode,
                "sparse": self.sparse,
            }
        )
        return config

    def _check_at_least_two_inputs(self, inputs):
        if not isinstance(inputs, (list, tuple)):
            raise ValueError(
                "`HashedCrossing` should be called on a list or tuple of "
                f"inputs. Received: inputs={inputs}"
            )
        if len(inputs) < 2:
            raise ValueError(
                "`HashedCrossing` should be called on at least two inputs. "
                f"Received: inputs={inputs}"
            )

    def _check_input_shape_and_type(self, inputs):
        first_shape = inputs[0].shape.as_list()
        rank = len(first_shape)
        if rank > 2 or (rank == 2 and first_shape[-1] != 1):
            raise ValueError(
                "All `HashedCrossing` inputs should have shape `[]`, "
                "`[batch_size]` or `[batch_size, 1]`. "
                f"Received: inputs={inputs}"
            )
        if not all(x.shape.as_list() == first_shape for x in inputs[1:]):
            raise ValueError(
                "All `HashedCrossing` inputs should have equal shape. "
                f"Received: inputs={inputs}"
            )
        if any(
            isinstance(x, (tf.RaggedTensor, tf.SparseTensor)) for x in inputs
        ):
            raise ValueError(
                "All `HashedCrossing` inputs should be dense tensors. "
                f"Received: inputs={inputs}"
            )
        if not all(x.dtype.is_integer or x.dtype == tf.string for x in inputs):
            raise ValueError(
                "All `HashedCrossing` inputs should have an integer or "
                f"string dtype. Received: inputs={inputs}"
            )




# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities related to distributed training."""

import contextlib

import tensorflow.compat.v2 as tf
from absl import flags

from keras.src import backend

FLAGS = flags.FLAGS


# TODO(b/118776054): Currently we support global batch size for TPUStrategy and
# core MirroredStrategy only. Remove this check when contrib MirroredStrategy is
# no longer needed.
def global_batch_size_supported(distribution_strategy):
    return distribution_strategy.extended._global_batch_size


def call_replica_local_fn(fn, *args, **kwargs):
    """Call a function that uses replica-local variables.

    This function correctly handles calling `fn` in a cross-replica
    context.

    Args:
      fn: The function to call.
      *args: Positional arguments to the `fn`.
      **kwargs: Keyword argument to `fn`.

    Returns:
      The result of calling `fn`.
    """
    # TODO(b/132666209): Remove this function when we support assign_*
    # for replica-local variables.
    strategy = None
    if "strategy" in kwargs:
        strategy = kwargs.pop("strategy")
    else:
        if tf.distribute.has_strategy():
            strategy = tf.distribute.get_strategy()

    # TODO(b/120571621): TPUStrategy does not implement replica-local variables.
    is_tpu = backend.is_tpu_strategy(strategy)
    if (not is_tpu) and strategy and tf.distribute.in_cross_replica_context():
        with strategy.scope():
            return strategy.extended.call_for_each_replica(fn, args, kwargs)
    return fn(*args, **kwargs)


def is_distributed_variable(v):
    """Returns whether `v` is a distributed variable."""
    return isinstance(v, tf.distribute.DistributedValues) and isinstance(
        v, tf.Variable
    )


def get_strategy():
    """Creates a `tf.distribute.Strategy` object from flags.

    Example usage:

    ```python
    strategy = utils.get_strategy()
    with strategy.scope():
      model = tf.keras.Sequential([tf.keras.layers.Dense(10)])

    model.compile(...)
    train_ds, test_ds = ...
    model.fit(train_ds, validation_data=test_ds, epochs=10)
    ```

    Returns:
      `tf.distribute.Strategy` instance.
    """
    cls = FLAGS.keras_distribute_strategy_class
    accepted_strats = {
        "tpu",
        "multi_worker_mirrored",
        "mirrored",
        "parameter_server",
        "one_device",
    }
    if cls == "tpu":
        tpu_addr = FLAGS.keras_distribute_strategy_tpu_addr
        if not tpu_addr:
            raise ValueError(
                "When using a TPU strategy, you must set the flag "
                "`keras_distribute_strategy_tpu_addr` (TPU address)."
            )
        cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
            tpu=tpu_addr
        )
        tf.config.experimental_connect_to_cluster(cluster_resolver)
        tf.tpu.experimental.initialize_tpu_system(cluster_resolver)
        strategy = tf.distribute.experimental.TPUStrategy(cluster_resolver)
    elif cls == "multi_worker_mirrored":
        strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy()
    elif cls == "mirrored":
        strategy = tf.distribute.MirroredStrategy()
    elif cls == "parameter_server":
        cluster_resolver = (
            tf.distribute.cluster_resolver.TFConfigClusterResolver()
        )
        strategy = tf.distribute.experimental.ParameterServerStrategy(
            cluster_resolver
        )
    elif cls == "one_device":
        strategy = tf.distribute.OneDeviceStrategy("/gpu:0")
    else:
        raise ValueError(
            "Unknown distribution strategy flag. Received: "
            f"keras_distribute_strategy_class={cls}. "
            f"It should be one of {accepted_strats}"
        )
    return strategy


def maybe_preemption_handler_scope(model):

    if getattr(model, "_preemption_handler", None):
        preemption_checkpoint_scope = (
            model._preemption_handler.watch_preemption_scope()
        )
    else:
        preemption_checkpoint_scope = contextlib.nullcontext()

    return preemption_checkpoint_scope




# Copyright 2022 The Keras Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Group normalization layer"""

import tensorflow.compat.v2 as tf

from keras.src import backend
from keras.src import constraints
from keras.src import initializers
from keras.src import regularizers
from keras.src.layers import InputSpec
from keras.src.layers import Layer
from keras.src.utils import tf_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.GroupNormalization", v1=[])
class GroupNormalization(Layer):
    """Group normalization layer.

    Group Normalization divides the channels into groups and computes
    within each group the mean and variance for normalization.
    Empirically, its accuracy is more stable than batch norm in a wide
    range of small batch sizes, if learning rate is adjusted linearly
    with batch sizes.

    Relation to Layer Normalization:
    If the number of groups is set to 1, then this operation becomes nearly
    identical to Layer Normalization (see Layer Normalization docs for details).

    Relation to Instance Normalization:
    If the number of groups is set to the input dimension (number of groups is
    equal to number of channels), then this operation becomes identical to
    Instance Normalization.

    Args:
      groups: Integer, the number of groups for Group Normalization. Can be in
        the range [1, N] where N is the input dimension. The input dimension
        must be divisible by the number of groups. Defaults to `32`.
      axis: Integer or List/Tuple. The axis or axes to normalize across.
        Typically, this is the features axis/axes. The left-out axes are
        typically the batch axis/axes. `-1` is the last dimension in the
        input. Defaults to `-1`.
      epsilon: Small float added to variance to avoid dividing by zero. Defaults
        to 1e-3
      center: If True, add offset of `beta` to normalized tensor. If False,
        `beta` is ignored. Defaults to `True`.
      scale: If True, multiply by `gamma`. If False, `gamma` is not used.
        When the next layer is linear (also e.g. `nn.relu`), this can be
        disabled since the scaling will be done by the next layer.
        Defaults to `True`.
      beta_initializer: Initializer for the beta weight. Defaults to zeros.
      gamma_initializer: Initializer for the gamma weight. Defaults to ones.
      beta_regularizer: Optional regularizer for the beta weight. None by
        default.
      gamma_regularizer: Optional regularizer for the gamma weight. None by
        default.
      beta_constraint: Optional constraint for the beta weight. None by default.
      gamma_constraint: Optional constraint for the gamma weight. None by
        default.  Input shape: Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis) when using this
        layer as the first layer in a model.  Output shape: Same shape as input.

    Call arguments:
      inputs: Input tensor (of any rank).
      mask: The mask parameter is a tensor that indicates the weight for each
        position in the input tensor when computing the mean and variance.

    Reference: - [Yuxin Wu & Kaiming He, 2018](https://arxiv.org/abs/1803.08494)
    """

    def __init__(
        self,
        groups=32,
        axis=-1,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.supports_masking = True
        self.groups = groups
        self.axis = axis
        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_constraint = constraints.get(beta_constraint)
        self.gamma_constraint = constraints.get(gamma_constraint)

    def build(self, input_shape):
        tf_utils.validate_axis(self.axis, input_shape)

        dim = input_shape[self.axis]
        if dim is None:
            raise ValueError(
                f"Axis {self.axis} of input tensor should have a defined "
                "dimension but the layer received an input with shape "
                f"{input_shape}."
            )

        if self.groups == -1:
            self.groups = dim

        if dim < self.groups:
            raise ValueError(
                f"Number of groups ({self.groups}) cannot be more than the "
                f"number of channels ({dim})."
            )

        if dim % self.groups != 0:
            raise ValueError(
                f"Number of groups ({self.groups}) must be a multiple "
                f"of the number of channels ({dim})."
            )

        self.input_spec = InputSpec(
            ndim=len(input_shape), axes={self.axis: dim}
        )

        if self.scale:
            self.gamma = self.add_weight(
                shape=(dim,),
                name="gamma",
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
            )
        else:
            self.gamma = None

        if self.center:
            self.beta = self.add_weight(
                shape=(dim,),
                name="beta",
                initializer=self.beta_initializer,
                regularizer=self.beta_regularizer,
                constraint=self.beta_constraint,
            )
        else:
            self.beta = None

        super().build(input_shape)

    def call(self, inputs, mask=None):
        input_shape = tf.shape(inputs)

        if mask is None:
            mask = tf.ones_like(inputs)
        else:
            # We broadcast before we group in case the mask does not have the
            # same shape as the input.
            mask = tf.broadcast_to(mask, input_shape)

        reshaped_inputs = self._reshape_into_groups(inputs)
        reshaped_mask = self._reshape_into_groups(mask)

        normalized_inputs = self._apply_normalization(
            reshaped_inputs=reshaped_inputs,
            input_shape=input_shape,
            reshaped_mask=reshaped_mask,
        )

        return tf.reshape(normalized_inputs, input_shape)

    def _reshape_into_groups(self, inputs):
        input_shape = tf.shape(inputs)
        group_shape = [input_shape[i] for i in range(inputs.shape.rank)]

        group_shape[self.axis] = input_shape[self.axis] // self.groups
        group_shape.insert(self.axis, self.groups)
        group_shape = tf.stack(group_shape)
        reshaped_inputs = tf.reshape(inputs, group_shape)
        return reshaped_inputs

    def _apply_normalization(
        self,
        *,
        reshaped_inputs,
        reshaped_mask,
        input_shape,
    ):
        group_reduction_axes = list(range(1, reshaped_inputs.shape.rank))

        axis = self.axis - 1
        group_reduction_axes.pop(axis)

        mask_weights = tf.cast(reshaped_mask, reshaped_inputs.dtype)

        mean, variance = tf.nn.weighted_moments(
            reshaped_inputs,
            axes=group_reduction_axes,
            frequency_weights=mask_weights,
            keepdims=True,
        )

        gamma, beta = self._get_reshaped_weights(input_shape)
        normalized_inputs = tf.nn.batch_normalization(
            reshaped_inputs,
            mean=mean,
            variance=variance,
            scale=gamma,
            offset=beta,
            variance_epsilon=self.epsilon,
        )
        return normalized_inputs

    def _get_reshaped_weights(self, input_shape):
        broadcast_shape = self._create_broadcast_shape(input_shape)
        gamma = None
        beta = None
        if self.scale:
            gamma = tf.reshape(self.gamma, broadcast_shape)

        if self.center:
            beta = tf.reshape(self.beta, broadcast_shape)
        return gamma, beta

    def _create_broadcast_shape(self, input_shape):
        broadcast_shape = [1] * backend.int_shape(input_shape)[0]

        broadcast_shape[self.axis] = input_shape[self.axis] // self.groups
        broadcast_shape.insert(self.axis, self.groups)

        return broadcast_shape

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = {
            "groups": self.groups,
            "axis": self.axis,
            "epsilon": self.epsilon,
            "center": self.center,
            "scale": self.scale,
            "beta_initializer": initializers.serialize(self.beta_initializer),
            "gamma_initializer": initializers.serialize(self.gamma_initializer),
            "beta_regularizer": regularizers.serialize(self.beta_regularizer),
            "gamma_regularizer": regularizers.serialize(self.gamma_regularizer),
            "beta_constraint": constraints.serialize(self.beta_constraint),
            "gamma_constraint": constraints.serialize(self.gamma_constraint),
        }
        base_config = super().get_config()
        return {**base_config, **config}




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the flatten layer."""


import functools
import operator

import numpy as np
import tensorflow.compat.v2 as tf

from keras.src.engine.base_layer import Layer
from keras.src.engine.input_spec import InputSpec
from keras.src.utils import conv_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.Flatten")
class Flatten(Layer):
    """Flattens the input. Does not affect the batch size.

    Note: If inputs are shaped `(batch,)` without a feature axis, then
    flattening adds an extra channel dimension and output shape is `(batch, 1)`.

    Args:
      data_format: A string,
        one of `channels_last` (default) or `channels_first`.
        The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch, ..., channels)` while `channels_first` corresponds to
        inputs with shape `(batch, channels, ...)`.
        When unspecified, uses
        `image_data_format` value found in your Keras config file at
         `~/.keras/keras.json` (if exists) else 'channels_last'.
        Defaults to 'channels_last'.

    Example:

    >>> model = tf.keras.Sequential()
    >>> model.add(tf.keras.layers.Conv2D(64, 3, 3, input_shape=(3, 32, 32)))
    >>> model.output_shape
    (None, 1, 10, 64)

    >>> model.add(Flatten())
    >>> model.output_shape
    (None, 640)

    """

    def __init__(self, data_format=None, **kwargs):
        super().__init__(**kwargs)
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.input_spec = InputSpec(min_ndim=1)
        self._channels_first = self.data_format == "channels_first"

    def call(self, inputs):
        if self._channels_first:
            rank = inputs.shape.rank
            if rank and rank > 1:
                # Switch to channels-last format.
                permutation = [0]
                permutation.extend(range(2, rank))
                permutation.append(1)
                inputs = tf.transpose(inputs, perm=permutation)

        if tf.executing_eagerly():
            # Full static shape is guaranteed to be available.
            # Performance: Using `constant_op` is much faster than passing a
            # list.
            flattened_shape = tf.constant([inputs.shape[0], -1])
            return tf.reshape(inputs, flattened_shape)
        else:
            input_shape = inputs.shape
            rank = input_shape.rank
            if rank == 1:
                return tf.expand_dims(inputs, axis=1)
            else:
                batch_dim = tf.compat.dimension_value(input_shape[0])
                non_batch_dims = input_shape[1:]
                # Reshape in a way that preserves as much shape info as
                # possible.
                if non_batch_dims.is_fully_defined():
                    last_dim = int(
                        functools.reduce(operator.mul, non_batch_dims)
                    )
                    flattened_shape = tf.constant([-1, last_dim])
                elif batch_dim is not None:
                    flattened_shape = tf.constant([int(batch_dim), -1])
                else:
                    flattened_shape = [tf.shape(inputs)[0], -1]
                return tf.reshape(inputs, flattened_shape)

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()
        if not input_shape:
            output_shape = tf.TensorShape([1])
        else:
            output_shape = [input_shape[0]]
        if np.all(input_shape[1:]):
            output_shape += [np.prod(input_shape[1:], dtype=int)]
        else:
            output_shape += [None]
        return tf.TensorShape(output_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"data_format": self.data_format})
        return config




"""AUTOGENERATED. DO NOT EDIT."""

from keras.src.backend import disable_tf_random_generator
from keras.src.backend import enable_tf_random_generator
from keras.src.backend import is_tf_random_generator_enabled



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Adagrad optimizer implementation."""

import tensorflow.compat.v2 as tf

from keras.src import initializers
from keras.src.optimizers import optimizer
from keras.src.saving.object_registration import register_keras_serializable

# isort: off
from tensorflow.python.util.tf_export import keras_export


@register_keras_serializable()
@keras_export(
    "keras.optimizers.experimental.Adagrad",
    "keras.optimizers.Adagrad",
    "keras.dtensor.experimental.optimizers.Adagrad",
    v1=[],
)
class Adagrad(optimizer.Optimizer):
    r"""Optimizer that implements the Adagrad algorithm.

    Adagrad is an optimizer with parameter-specific learning rates,
    which are adapted relative to how frequently a parameter gets
    updated during training. The more updates a parameter receives,
    the smaller the updates.

    Args:
        learning_rate: Initial value for the learning rate:
            either a floating point value,
            or a `tf.keras.optimizers.schedules.LearningRateSchedule` instance.
            Defaults to 0.001. Note that `Adagrad` tends to benefit from higher
            initial learning rate values compared to other optimizers. To match
            the exact form in the original paper, use 1.0.
        initial_accumulator_value: Floating point value.
            Starting value for the accumulators (per-parameter momentum values).
            Must be non-negative.
        epsilon: Small floating point value used to maintain numerical
            stability.
        {{base_optimizer_keyword_args}}

    Reference:
        - [Duchi et al., 2011](
            http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf).
    """

    def __init__(
        self,
        learning_rate=0.001,
        initial_accumulator_value=0.1,
        epsilon=1e-7,
        weight_decay=None,
        clipnorm=None,
        clipvalue=None,
        global_clipnorm=None,
        use_ema=False,
        ema_momentum=0.99,
        ema_overwrite_frequency=None,
        jit_compile=True,
        name="Adagrad",
        **kwargs
    ):
        super().__init__(
            weight_decay=weight_decay,
            clipnorm=clipnorm,
            clipvalue=clipvalue,
            global_clipnorm=global_clipnorm,
            use_ema=use_ema,
            ema_momentum=ema_momentum,
            ema_overwrite_frequency=ema_overwrite_frequency,
            jit_compile=jit_compile,
            name=name,
            **kwargs
        )
        self._learning_rate = self._build_learning_rate(learning_rate)
        self.initial_accumulator_value = initial_accumulator_value
        self.epsilon = epsilon

    def build(self, var_list):
        super().build(var_list)
        if hasattr(self, "_built") and self._built:
            return
        self._built = True
        self._accumulators = []
        initializer = initializers.Constant(self.initial_accumulator_value)
        for var in var_list:
            self._accumulators.append(
                self.add_variable_from_reference(
                    var,
                    "accumulator",
                    initial_value=initializer(shape=var.shape, dtype=var.dtype),
                )
            )

    def update_step(self, grad, variable):
        """Update step given gradient and the associated model variable."""
        lr = tf.cast(self.learning_rate, variable.dtype)

        var_key = self._var_key(variable)
        accumulator = self._accumulators[self._index_dict[var_key]]

        if isinstance(grad, tf.IndexedSlices):
            # Sparse gradients.
            accumulator.scatter_add(
                tf.IndexedSlices(grad.values * grad.values, grad.indices)
            )
            sparse_accumulator = tf.gather(accumulator, indices=grad.indices)
            sparse_denominator = tf.sqrt(sparse_accumulator + self.epsilon)
            variable.scatter_add(
                tf.IndexedSlices(
                    -lr * grad.values / sparse_denominator, grad.indices
                )
            )
        else:
            # Dense gradients.
            accumulator.assign_add(grad * grad)
            variable.assign_sub(lr * grad / tf.sqrt(accumulator + self.epsilon))

    def get_config(self):
        config = super().get_config()

        config.update(
            {
                "learning_rate": self._serialize_hyperparameter(
                    self._learning_rate
                ),
                "initial_accumulator_value": self.initial_accumulator_value,
                "epsilon": self.epsilon,
            }
        )
        return config


Adagrad.__doc__ = Adagrad.__doc__.replace(
    "{{base_optimizer_keyword_args}}", optimizer.base_optimizer_keyword_args
)




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Softmax activation layer."""


import tensorflow.compat.v2 as tf

from keras.src import backend
from keras.src.engine.base_layer import Layer
from keras.src.utils import tf_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


def _large_compatible_negative(tensor_type):
    """Large negative number as Tensor.

    This function is necessary because the standard value for epsilon
    in this module (-1e9) cannot be represented using tf.float16

    Args:
        tensor_type: a dtype to determine the type.

    Returns:
        a large negative number.
    """
    # In case of dtype=float16 (e.g., for mixed-precision), the largest
    # negative number (dtypes.float16.min) is divided by 2, in order to
    # avoid overflows when summing negative inputs.
    if tensor_type == tf.float16:
        return tf.float16.min / 2.0
    return -1e9


@keras_export("keras.layers.Softmax")
class Softmax(Layer):
    """Softmax activation function.

    Example without mask:

    >>> inp = np.asarray([[1., 2., 1.]])
    >>> layer = tf.keras.layers.Softmax()
    >>> layer(inp).numpy()
    array([[0.21194157, 0.5761169 , 0.21194157]], dtype=float32)
    >>> mask = np.asarray([[True, False, True]], dtype=bool)
    >>> layer(inp, mask).numpy()
    array([[0.5, 0. , 0.5]], dtype=float32)

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as the input.

    Args:
        axis: Integer, or list of Integers, axis along which the softmax
            normalization is applied.
    Call arguments:
        inputs: The inputs, or logits to the softmax layer.
        mask: A boolean mask of the same shape as `inputs`. The mask
            specifies 1 to keep and 0 to mask. Defaults to `None`.


    Returns:
        Softmaxed output with the same shape as `inputs`.
    """

    def __init__(self, axis=-1, **kwargs):
        super().__init__(**kwargs)
        self.supports_masking = True
        self.axis = axis

    def call(self, inputs, mask=None):
        if mask is not None:
            # Since mask is 1.0 for positions we want to keep and 0.0 for masked
            # positions, this operation will create a tensor which is 0.0 for
            # positions we want to attend and -1e.9 for masked positions.
            adder = (1.0 - tf.cast(mask, inputs.dtype)) * (
                _large_compatible_negative(inputs.dtype)
            )

            # Since we are adding it to the raw scores before the softmax, this
            # is effectively the same as removing these entirely.
            inputs += adder
        if isinstance(self.axis, (tuple, list)):
            if len(self.axis) > 1:
                return tf.exp(
                    inputs
                    - tf.reduce_logsumexp(inputs, axis=self.axis, keepdims=True)
                )
            else:
                return backend.softmax(inputs, axis=self.axis[0])
        return backend.softmax(inputs, axis=self.axis)

    def get_config(self):
        config = {"axis": self.axis}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @tf_utils.shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape




# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras estimator API."""

import tensorflow.compat.v2 as tf

# isort: off
from tensorflow.python.util.tf_export import keras_export

# Keras has undeclared dependency on tensorflow/estimator:estimator_py.
# As long as you depend //third_party/py/tensorflow:tensorflow target
# everything will work as normal.

_model_to_estimator_usage_gauge = tf.__internal__.monitoring.BoolGauge(
    "/tensorflow/api/keras/model_to_estimator",
    "Whether tf.keras.estimator.model_to_estimator() is called.",
    "version",
)


# LINT.IfChange
@keras_export(v1=["keras.estimator.model_to_estimator"])
def model_to_estimator(
    keras_model=None,
    keras_model_path=None,
    custom_objects=None,
    model_dir=None,
    config=None,
    checkpoint_format="saver",
    metric_names_map=None,
    export_outputs=None,
):
    """Constructs an `Estimator` instance from given keras model.

    If you use infrastructure or other tooling that relies on Estimators, you
    can still build a Keras model and use model_to_estimator to convert the
    Keras model to an Estimator for use with downstream systems.

    For usage example, please see:
    [Creating estimators from Keras Models](
    https://www.tensorflow.org/guide/estimator#create_an_estimator_from_a_keras_model).

    Sample Weights:
    Estimators returned by `model_to_estimator` are configured so that they can
    handle sample weights (similar to `keras_model.fit(x, y, sample_weights)`).

    To pass sample weights when training or evaluating the Estimator, the first
    item returned by the input function should be a dictionary with keys
    `features` and `sample_weights`. Example below:

    ```python
    keras_model = tf.keras.Model(...)
    keras_model.compile(...)

    estimator = tf.keras.estimator.model_to_estimator(keras_model)

    def input_fn():
      return dataset_ops.Dataset.from_tensors(
          ({'features': features, 'sample_weights': sample_weights},
           targets))

    estimator.train(input_fn, steps=1)
    ```

    Example with customized export signature:
    ```python
    inputs = {'a': tf.keras.Input(..., name='a'),
              'b': tf.keras.Input(..., name='b')}
    outputs = {'c': tf.keras.layers.Dense(..., name='c')(inputs['a']),
               'd': tf.keras.layers.Dense(..., name='d')(inputs['b'])}
    keras_model = tf.keras.Model(inputs, outputs)
    keras_model.compile(...)
    export_outputs = {'c': tf.estimator.export.RegressionOutput,
                      'd': tf.estimator.export.ClassificationOutput}

    estimator = tf.keras.estimator.model_to_estimator(
        keras_model, export_outputs=export_outputs)

    def input_fn():
      return dataset_ops.Dataset.from_tensors(
          ({'features': features, 'sample_weights': sample_weights},
           targets))

    estimator.train(input_fn, steps=1)
    ```

    Args:
      keras_model: A compiled Keras model object. This argument is mutually
        exclusive with `keras_model_path`. Estimator's `model_fn` uses the
        structure of the model to clone the model. Defaults to `None`.
      keras_model_path: Path to a compiled Keras model saved on disk, in HDF5
        format, which can be generated with the `save()` method of a Keras
        model.  This argument is mutually exclusive with `keras_model`.
        Defaults to `None`.
      custom_objects: Dictionary for cloning customized objects. This is
        used with classes that is not part of this pip package. For example, if
        user maintains a `relu6` class that inherits from
        `tf.keras.layers.Layer`, then pass `custom_objects={'relu6': relu6}`.
        Defaults to `None`.
      model_dir: Directory to save `Estimator` model parameters, graph, summary
        files for TensorBoard, etc. If unset a directory will be created with
        `tempfile.mkdtemp`
      config: `RunConfig` to config `Estimator`. Allows setting up things in
        `model_fn` based on configuration such as `num_ps_replicas`, or
        `model_dir`. If both `config.model_dir` and the
        `model_dir` argument (above) are specified the `model_dir` **argument**
        takes precedence. Defaults to `None`.
      checkpoint_format: Sets the format of the checkpoint saved by the
        estimator when training. May be `saver` or `checkpoint`, depending on
        whether to save checkpoints from `tf.train.Saver` or
        `tf.train.Checkpoint`. Estimators use name-based `tf.train.Saver`
        checkpoints, while Keras models use object-based checkpoints from
        `tf.train.Checkpoint`. Currently, saving object-based checkpoints
        from `model_to_estimator` is only supported by Functional and
        Sequential models. Defaults to 'saver'.
      metric_names_map: Optional dictionary mapping Keras model output metric
        names to custom names. This can be used to override the default Keras
        model output metrics names in a multi IO model use case and provide
        custom names for the `eval_metric_ops` in Estimator.
        The Keras model metric names can be obtained using `model.metrics_names`
        excluding any loss metrics such as total loss and output losses.
        For example, if your Keras model has two outputs `out_1` and `out_2`,
        with `mse` loss and `acc` metric, then `model.metrics_names` will be
        `['loss', 'out_1_loss', 'out_2_loss', 'out_1_acc', 'out_2_acc']`.
        The model metric names excluding the loss metrics will be
        `['out_1_acc', 'out_2_acc']`.
      export_outputs: Optional dictionary. This can be used to override the
        default Keras model output exports in a multi IO model use case and
        provide custom names for the `export_outputs` in
        `tf.estimator.EstimatorSpec`. Default is None, which is equivalent to
        {'serving_default': `tf.estimator.export.PredictOutput`}. If not None,
        the keys must match the keys of `model.output_names`.
        A dict `{name: output}` where:
          * name: An arbitrary name for this output.
          * output: an `ExportOutput` class such as `ClassificationOutput`,
            `RegressionOutput`, or `PredictOutput`. Single-headed models only
            need to specify one entry in this dictionary. Multi-headed models
            should specify one entry for each head, one of which must be named
            using
            `tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY`
            If no entry is provided, a default `PredictOutput` mapping to
            `predictions` will be created.

    Returns:
      An Estimator from given keras model.

    Raises:
      ValueError: If neither keras_model nor keras_model_path was given.
      ValueError: If both keras_model and keras_model_path was given.
      ValueError: If the keras_model_path is a GCS URI.
      ValueError: If keras_model has not been compiled.
      ValueError: If an invalid checkpoint_format was given.
    """

    try:
        # isort: off
        from tensorflow_estimator.python.estimator import (
            keras_lib,
        )
    except ImportError:
        raise NotImplementedError(
            "tf.keras.estimator.model_to_estimator function not available in "
            "your installation."
        )
    _model_to_estimator_usage_gauge.get_cell("v1").set(True)
    return keras_lib.model_to_estimator(
        keras_model=keras_model,
        keras_model_path=keras_model_path,
        custom_objects=custom_objects,
        model_dir=model_dir,
        config=config,
        checkpoint_format=checkpoint_format,
        use_v2_estimator=False,
        metric_names_map=metric_names_map,
        export_outputs=export_outputs,
    )


@keras_export("keras.estimator.model_to_estimator", v1=[])
def model_to_estimator_v2(
    keras_model=None,
    keras_model_path=None,
    custom_objects=None,
    model_dir=None,
    config=None,
    checkpoint_format="checkpoint",
    metric_names_map=None,
    export_outputs=None,
):
    """Constructs an `Estimator` instance from given keras model.

    If you use infrastructure or other tooling that relies on Estimators, you
    can still build a Keras model and use model_to_estimator to convert the
    Keras model to an Estimator for use with downstream systems.

    For usage example, please see:
    [Creating estimators from Keras Models](
    https://www.tensorflow.org/guide/estimators#creating_estimators_from_keras_models).

    Sample Weights:
    Estimators returned by `model_to_estimator` are configured so that they can
    handle sample weights (similar to `keras_model.fit(x, y, sample_weights)`).

    To pass sample weights when training or evaluating the Estimator, the first
    item returned by the input function should be a dictionary with keys
    `features` and `sample_weights`. Example below:

    ```python
    keras_model = tf.keras.Model(...)
    keras_model.compile(...)

    estimator = tf.keras.estimator.model_to_estimator(keras_model)

    def input_fn():
      return dataset_ops.Dataset.from_tensors(
          ({'features': features, 'sample_weights': sample_weights},
           targets))

    estimator.train(input_fn, steps=1)
    ```

    Example with customized export signature:
    ```python
    inputs = {'a': tf.keras.Input(..., name='a'),
              'b': tf.keras.Input(..., name='b')}
    outputs = {'c': tf.keras.layers.Dense(..., name='c')(inputs['a']),
               'd': tf.keras.layers.Dense(..., name='d')(inputs['b'])}
    keras_model = tf.keras.Model(inputs, outputs)
    keras_model.compile(...)
    export_outputs = {'c': tf.estimator.export.RegressionOutput,
                      'd': tf.estimator.export.ClassificationOutput}

    estimator = tf.keras.estimator.model_to_estimator(
        keras_model, export_outputs=export_outputs)

    def input_fn():
      return dataset_ops.Dataset.from_tensors(
          ({'features': features, 'sample_weights': sample_weights},
           targets))

    estimator.train(input_fn, steps=1)
    ```

    Note: We do not support creating weighted metrics in Keras and converting
    them to weighted metrics in the Estimator API using `model_to_estimator`.
    You will have to create these metrics directly on the estimator spec using
    the `add_metrics` function.

    To customize the estimator `eval_metric_ops` names, you can pass in the
    `metric_names_map` dictionary mapping the keras model output metric names
    to the custom names as follows:

    ```python
      input_a = tf.keras.layers.Input(shape=(16,), name='input_a')
      input_b = tf.keras.layers.Input(shape=(16,), name='input_b')
      dense = tf.keras.layers.Dense(8, name='dense_1')
      interm_a = dense(input_a)
      interm_b = dense(input_b)
      merged = tf.keras.layers.concatenate([interm_a, interm_b], name='merge')
      output_a = tf.keras.layers.Dense(3, activation='softmax', name='dense_2')(
              merged)
      output_b = tf.keras.layers.Dense(2, activation='softmax', name='dense_3')(
              merged)
      keras_model = tf.keras.models.Model(
          inputs=[input_a, input_b], outputs=[output_a, output_b])
      keras_model.compile(
          loss='categorical_crossentropy',
          optimizer='rmsprop',
          metrics={
              'dense_2': 'categorical_accuracy',
              'dense_3': 'categorical_accuracy'
          })

      metric_names_map = {
          'dense_2_categorical_accuracy': 'acc_1',
          'dense_3_categorical_accuracy': 'acc_2',
      }
      keras_est = tf.keras.estimator.model_to_estimator(
          keras_model=keras_model,
          config=config,
          metric_names_map=metric_names_map)
    ```

    Args:
      keras_model: A compiled Keras model object. This argument is mutually
        exclusive with `keras_model_path`. Estimator's `model_fn` uses the
        structure of the model to clone the model. Defaults to `None`.
      keras_model_path: Path to a compiled Keras model saved on disk, in HDF5
        format, which can be generated with the `save()` method of a Keras
        model.  This argument is mutually exclusive with `keras_model`.
        Defaults to `None`.
      custom_objects: Dictionary for cloning customized objects. This is
        used with classes that is not part of this pip package. For example, if
        user maintains a `relu6` class that inherits from
        `tf.keras.layers.Layer`, then pass `custom_objects={'relu6': relu6}`.
        Defaults to `None`.
      model_dir: Directory to save `Estimator` model parameters, graph, summary
        files for TensorBoard, etc. If unset a directory will be created with
        `tempfile.mkdtemp`
      config: `RunConfig` to config `Estimator`. Allows setting up things in
        `model_fn` based on configuration such as `num_ps_replicas`, or
        `model_dir`. If both `config.model_dir` and the
        `model_dir` argument (above) are specified the `model_dir` **argument**
        takes precedence. Defaults to `None`.
      checkpoint_format: Sets the format of the checkpoint saved by the
        estimator when training. May be `saver` or `checkpoint`, depending on
        whether to save checkpoints from `tf.compat.v1.train.Saver` or
        `tf.train.Checkpoint`.  The default is `checkpoint`. Estimators use
        name-based `tf.train.Saver` checkpoints, while Keras models use
        object-based checkpoints from `tf.train.Checkpoint`. Currently, saving
        object-based checkpoints from `model_to_estimator` is only supported by
        Functional and Sequential models. Defaults to 'checkpoint'.
      metric_names_map: Optional dictionary mapping Keras model output metric
        names to custom names. This can be used to override the default Keras
        model output metrics names in a multi IO model use case and provide
        custom names for the `eval_metric_ops` in Estimator.
        The Keras model metric names can be obtained using `model.metrics_names`
        excluding any loss metrics such as total loss and output losses.
        For example, if your Keras model has two outputs `out_1` and `out_2`,
        with `mse` loss and `acc` metric, then `model.metrics_names` will be
        `['loss', 'out_1_loss', 'out_2_loss', 'out_1_acc', 'out_2_acc']`.
        The model metric names excluding the loss metrics will be
        `['out_1_acc', 'out_2_acc']`.
      export_outputs: Optional dictionary. This can be used to override the
        default Keras model output exports in a multi IO model use case and
        provide custom names for the `export_outputs` in
        `tf.estimator.EstimatorSpec`. Default is None, which is equivalent to
        {'serving_default': `tf.estimator.export.PredictOutput`}. If not None,
        the keys must match the keys of `model.output_names`.
        A dict `{name: output}` where:
          * name: An arbitrary name for this output.
          * output: an `ExportOutput` class such as `ClassificationOutput`,
            `RegressionOutput`, or `PredictOutput`. Single-headed models only
            need to specify one entry in this dictionary. Multi-headed models
            should specify one entry for each head, one of which must be named
            using
            `tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY`
            If no entry is provided, a default `PredictOutput` mapping to
            `predictions` will be created.

    Returns:
      An Estimator from given keras model.

    Raises:
      ValueError: If neither keras_model nor keras_model_path was given.
      ValueError: If both keras_model and keras_model_path was given.
      ValueError: If the keras_model_path is a GCS URI.
      ValueError: If keras_model has not been compiled.
      ValueError: If an invalid checkpoint_format was given.
    """

    try:
        # isort: off
        from tensorflow_estimator.python.estimator import (
            keras_lib,
        )
    except ImportError:
        raise NotImplementedError(
            "tf.keras.estimator.model_to_estimator function not available in "
            "your installation."
        )
    _model_to_estimator_usage_gauge.get_cell("v2").set(True)
    return keras_lib.model_to_estimator(
        keras_model=keras_model,
        keras_model_path=keras_model_path,
        custom_objects=custom_objects,
        model_dir=model_dir,
        config=config,
        checkpoint_format=checkpoint_format,
        use_v2_estimator=True,
        metric_names_map=metric_names_map,
        export_outputs=export_outputs,
    )


# LINT.ThenChange(//tensorflow_estimator/python/estimator/keras_lib.py)




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras abstract base layer for separable nD convolution."""


import tensorflow.compat.v2 as tf

from keras.src import activations
from keras.src import constraints
from keras.src import initializers
from keras.src import regularizers
from keras.src.engine.input_spec import InputSpec
from keras.src.layers.convolutional.base_conv import Conv


class SeparableConv(Conv):
    """Abstract base layer for separable nD convolution.

    This layer performs a depthwise convolution that acts separately on
    channels, followed by a pointwise convolution that mixes channels.
    If `use_bias` is True and a bias initializer is provided,
    it adds a bias vector to the output.
    It then optionally applies an activation function to produce the final
    output.

    Args:
      rank: An integer, the rank of the convolution, e.g. "2" for 2D
        convolution.
      filters: Integer, the dimensionality of the output space (i.e. the number
        of filters in the convolution).
      kernel_size: A tuple or list of integers specifying the spatial
        dimensions of the filters. Can be a single integer to specify the same
        value for all spatial dimensions.
      strides: A tuple or list of integers specifying the strides
        of the convolution. Can be a single integer to specify the same value
        for all spatial dimensions.
        Specifying any `stride` value != 1 is incompatible with specifying
        any `dilation_rate` value != 1.
      padding: One of `"valid"` or `"same"` (case-insensitive).
        `"valid"` means no padding. `"same"` results in padding with zeros
        evenly to the left/right or up/down of the input such that output has
        the same height/width dimension as the input.
      data_format: A string, one of `channels_last` (default) or
        `channels_first`.  The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch_size, ..., channels)` while `channels_first` corresponds to
        inputs with shape `(batch_size, channels, ...)`.
      dilation_rate: An integer or tuple/list of 2 integers, specifying
        the dilation rate to use for dilated convolution.
        Can be a single integer to specify the same value for
        all spatial dimensions.
        Currently, specifying any `dilation_rate` value != 1 is
        incompatible with specifying any stride value != 1.
      depth_multiplier: The number of depthwise convolution output channels for
        each input channel. The total number of depthwise convolution output
        channels will be equal to `num_filters_in * depth_multiplier`.
      activation: Activation function to use.
        If you don't specify anything, no activation is applied
        (see `keras.activations`).
      use_bias: Boolean, whether the layer uses a bias.
      depthwise_initializer: An initializer for the depthwise convolution kernel
        (see `keras.initializers`). If None, then the default initializer
        ('glorot_uniform') will be used.
      pointwise_initializer: An initializer for the pointwise convolution kernel
        (see `keras.initializers`). If None, then the default initializer
        ('glorot_uniform') will be used.
      bias_initializer: An initializer for the bias vector. If None, the default
        initializer ('zeros') will be used (see `keras.initializers`).
      depthwise_regularizer: Optional regularizer for the depthwise
        convolution kernel.
      pointwise_regularizer: Optional regularizer for the pointwise
        convolution kernel.
      bias_regularizer: Optional regularizer for the bias vector.
      activity_regularizer: Optional regularizer function for the output.
      depthwise_constraint: Optional projection function to be applied to the
        depthwise kernel after being updated by an `Optimizer` (e.g. used for
        norm constraints or value constraints for layer weights). The function
        must take as input the unprojected variable and must return the
        projected variable (which must have the same shape). Constraints are
        not safe to use when doing asynchronous distributed training.
      pointwise_constraint: Optional projection function to be applied to the
        pointwise kernel after being updated by an `Optimizer`.
      bias_constraint: Optional projection function to be applied to the
        bias after being updated by an `Optimizer`.
      trainable: Boolean, if `True` the weights of this layer will be marked as
        trainable (and listed in `layer.trainable_weights`).
    """

    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        depth_multiplier=1,
        activation=None,
        use_bias=True,
        depthwise_initializer="glorot_uniform",
        pointwise_initializer="glorot_uniform",
        bias_initializer="zeros",
        depthwise_regularizer=None,
        pointwise_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        depthwise_constraint=None,
        pointwise_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        **kwargs,
    ):
        super().__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=activations.get(activation),
            use_bias=use_bias,
            bias_initializer=initializers.get(bias_initializer),
            bias_regularizer=regularizers.get(bias_regularizer),
            activity_regularizer=regularizers.get(activity_regularizer),
            bias_constraint=bias_constraint,
            trainable=trainable,
            name=name,
            **kwargs,
        )
        self.depth_multiplier = depth_multiplier
        self.depthwise_initializer = initializers.get(depthwise_initializer)
        self.pointwise_initializer = initializers.get(pointwise_initializer)
        self.depthwise_regularizer = regularizers.get(depthwise_regularizer)
        self.pointwise_regularizer = regularizers.get(pointwise_regularizer)
        self.depthwise_constraint = constraints.get(depthwise_constraint)
        self.pointwise_constraint = constraints.get(pointwise_constraint)

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        channel_axis = self._get_channel_axis()
        if input_shape.dims[channel_axis].value is None:
            raise ValueError(
                "The channel dimension of the inputs should be defined. "
                f"The input_shape received is {input_shape}, "
                f"where axis {channel_axis} (0-based) "
                "is the channel dimension, which found to be `None`."
            )
        input_dim = int(input_shape[channel_axis])
        self.input_spec = InputSpec(
            ndim=self.rank + 2, axes={channel_axis: input_dim}
        )
        depthwise_kernel_shape = self.kernel_size + (
            input_dim,
            self.depth_multiplier,
        )
        pointwise_kernel_shape = (1,) * self.rank + (
            self.depth_multiplier * input_dim,
            self.filters,
        )

        self.depthwise_kernel = self.add_weight(
            name="depthwise_kernel",
            shape=depthwise_kernel_shape,
            initializer=self.depthwise_initializer,
            regularizer=self.depthwise_regularizer,
            constraint=self.depthwise_constraint,
            trainable=True,
            dtype=self.dtype,
        )
        self.pointwise_kernel = self.add_weight(
            name="pointwise_kernel",
            shape=pointwise_kernel_shape,
            initializer=self.pointwise_initializer,
            regularizer=self.pointwise_regularizer,
            constraint=self.pointwise_constraint,
            trainable=True,
            dtype=self.dtype,
        )
        if self.use_bias:
            self.bias = self.add_weight(
                name="bias",
                shape=(self.filters,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype,
            )
        else:
            self.bias = None
        self.built = True

    def call(self, inputs):
        raise NotImplementedError

    def get_config(self):
        config = {
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "padding": self.padding,
            "data_format": self.data_format,
            "depth_multiplier": self.depth_multiplier,
            "dilation_rate": self.dilation_rate,
            "activation": activations.serialize(self.activation),
            "use_bias": self.use_bias,
            "depthwise_initializer": initializers.serialize(
                self.depthwise_initializer
            ),
            "pointwise_initializer": initializers.serialize(
                self.pointwise_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "depthwise_regularizer": regularizers.serialize(
                self.depthwise_regularizer
            ),
            "pointwise_regularizer": regularizers.serialize(
                self.pointwise_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "activity_regularizer": regularizers.serialize(
                self.activity_regularizer
            ),
            "depthwise_constraint": constraints.serialize(
                self.depthwise_constraint
            ),
            "pointwise_constraint": constraints.serialize(
                self.pointwise_constraint
            ),
            "bias_constraint": constraints.serialize(self.bias_constraint),
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))




# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Benchmark for KPL implementation of categorical cross hash columns with dense
inputs."""


import tensorflow.compat.v2 as tf

import keras.src as keras
from keras.src.layers.preprocessing import hashed_crossing
from keras.src.layers.preprocessing.benchmarks import (
    feature_column_benchmark as fc_bm,
)

# isort: off
from tensorflow.python.eager.def_function import (
    function as tf_function,
)

NUM_REPEATS = 10
BATCH_SIZES = [32, 256]


def embedding_varlen(batch_size):
    """Benchmark a variable-length embedding."""
    # Data and constants.
    num_buckets = 10000
    data_a = tf.random.uniform(
        shape=(batch_size * NUM_REPEATS, 1), maxval=32768, dtype=tf.int64
    )
    data_b = tf.strings.as_string(data_a)

    # Keras implementation
    input_1 = keras.Input(shape=(1,), name="data_a", dtype=tf.int64)
    input_2 = keras.Input(shape=(1,), name="data_b", dtype=tf.string)
    outputs = hashed_crossing.HashedCrossing(num_buckets)([input_1, input_2])
    model = keras.Model([input_1, input_2], outputs)

    # FC implementation
    fc = tf.feature_column.crossed_column(["data_a", "data_b"], num_buckets)

    # Wrap the FC implementation in a tf.function for a fair comparison
    @tf_function()
    def fc_fn(tensors):
        fc.transform_feature(
            tf.__internal__.feature_column.FeatureTransformationCache(tensors),
            None,
        )

    # Benchmark runs
    keras_data = {
        "data_a": data_a,
        "data_b": data_b,
    }
    k_avg_time = fc_bm.run_keras(keras_data, model, batch_size, NUM_REPEATS)

    fc_data = {
        "data_a": data_a,
        "data_b": data_b,
    }
    fc_avg_time = fc_bm.run_fc(fc_data, fc_fn, batch_size, NUM_REPEATS)

    return k_avg_time, fc_avg_time


class BenchmarkLayer(fc_bm.LayerBenchmark):
    """Benchmark the layer forward pass."""

    def benchmark_layer(self):
        for batch in BATCH_SIZES:
            name = f"hashed_cross|dense|batch_{batch}"
            k_time, f_time = embedding_varlen(batch_size=batch)
            self.report(name, k_time, f_time, NUM_REPEATS)


if __name__ == "__main__":
    tf.test.main()




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras cropping layer for 3D input."""


import tensorflow.compat.v2 as tf

from keras.src.engine.base_layer import Layer
from keras.src.engine.input_spec import InputSpec
from keras.src.utils import conv_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.Cropping3D")
class Cropping3D(Layer):
    """Cropping layer for 3D data (e.g. spatial or spatio-temporal).

      Examples:

    >>> input_shape = (2, 28, 28, 10, 3)
    >>> x = np.arange(np.prod(input_shape)).reshape(input_shape)
    >>> y = tf.keras.layers.Cropping3D(cropping=(2, 4, 2))(x)
    >>> print(y.shape)
    (2, 24, 20, 6, 3)

    Args:
      cropping: Int, or tuple of 3 ints, or tuple of 3 tuples of 2 ints.
        - If int: the same symmetric cropping
          is applied to depth, height, and width.
        - If tuple of 3 ints: interpreted as two different
          symmetric cropping values for depth, height, and width:
          `(symmetric_dim1_crop, symmetric_dim2_crop, symmetric_dim3_crop)`.
        - If tuple of 3 tuples of 2 ints: interpreted as
          `((left_dim1_crop, right_dim1_crop), (left_dim2_crop,
            right_dim2_crop), (left_dim3_crop, right_dim3_crop))`
      data_format: A string,
        one of `channels_last` (default) or `channels_first`.
        The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch_size, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
        while `channels_first` corresponds to inputs with shape
        `(batch_size, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
        When unspecified, uses
        `image_data_format` value found in your Keras config file at
         `~/.keras/keras.json` (if exists) else 'channels_last'.
        Defaults to 'channels_last'.

    Input shape:
      5D tensor with shape:
      - If `data_format` is `"channels_last"`:
        `(batch_size, first_axis_to_crop, second_axis_to_crop,
        third_axis_to_crop, depth)`
      - If `data_format` is `"channels_first"`:
        `(batch_size, depth, first_axis_to_crop, second_axis_to_crop,
          third_axis_to_crop)`

    Output shape:
      5D tensor with shape:
      - If `data_format` is `"channels_last"`:
        `(batch_size, first_cropped_axis, second_cropped_axis,
        third_cropped_axis, depth)`
      - If `data_format` is `"channels_first"`:
        `(batch_size, depth, first_cropped_axis, second_cropped_axis,
          third_cropped_axis)`
    """

    def __init__(
        self, cropping=((1, 1), (1, 1), (1, 1)), data_format=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.data_format = conv_utils.normalize_data_format(data_format)
        if isinstance(cropping, int):
            self.cropping = (
                (cropping, cropping),
                (cropping, cropping),
                (cropping, cropping),
            )
        elif hasattr(cropping, "__len__"):
            if len(cropping) != 3:
                raise ValueError(
                    f"`cropping` should have 3 elements. Received: {cropping}."
                )
            dim1_cropping = conv_utils.normalize_tuple(
                cropping[0], 2, "1st entry of cropping", allow_zero=True
            )
            dim2_cropping = conv_utils.normalize_tuple(
                cropping[1], 2, "2nd entry of cropping", allow_zero=True
            )
            dim3_cropping = conv_utils.normalize_tuple(
                cropping[2], 2, "3rd entry of cropping", allow_zero=True
            )
            self.cropping = (dim1_cropping, dim2_cropping, dim3_cropping)
        else:
            raise ValueError(
                "`cropping` should be either an int, "
                "a tuple of 3 ints "
                "(symmetric_dim1_crop, symmetric_dim2_crop, "
                "symmetric_dim3_crop), "
                "or a tuple of 3 tuples of 2 ints "
                "((left_dim1_crop, right_dim1_crop),"
                " (left_dim2_crop, right_dim2_crop),"
                " (left_dim3_crop, right_dim2_crop)). "
                f"Received: {cropping}."
            )
        self.input_spec = InputSpec(ndim=5)

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()

        if self.data_format == "channels_first":
            if input_shape[2] is not None:
                dim1 = (
                    input_shape[2] - self.cropping[0][0] - self.cropping[0][1]
                )
            else:
                dim1 = None
            if input_shape[3] is not None:
                dim2 = (
                    input_shape[3] - self.cropping[1][0] - self.cropping[1][1]
                )
            else:
                dim2 = None
            if input_shape[4] is not None:
                dim3 = (
                    input_shape[4] - self.cropping[2][0] - self.cropping[2][1]
                )
            else:
                dim3 = None
            return tf.TensorShape(
                [input_shape[0], input_shape[1], dim1, dim2, dim3]
            )
        elif self.data_format == "channels_last":
            if input_shape[1] is not None:
                dim1 = (
                    input_shape[1] - self.cropping[0][0] - self.cropping[0][1]
                )
            else:
                dim1 = None
            if input_shape[2] is not None:
                dim2 = (
                    input_shape[2] - self.cropping[1][0] - self.cropping[1][1]
                )
            else:
                dim2 = None
            if input_shape[3] is not None:
                dim3 = (
                    input_shape[3] - self.cropping[2][0] - self.cropping[2][1]
                )
            else:
                dim3 = None
            return tf.TensorShape(
                [input_shape[0], dim1, dim2, dim3, input_shape[4]]
            )

    def call(self, inputs):

        if self.data_format == "channels_first":
            if (
                self.cropping[0][1]
                == self.cropping[1][1]
                == self.cropping[2][1]
                == 0
            ):
                return inputs[
                    :,
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] :,
                    self.cropping[2][0] :,
                ]
            elif self.cropping[0][1] == self.cropping[1][1] == 0:
                return inputs[
                    :,
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] :,
                    self.cropping[2][0] : -self.cropping[2][1],
                ]
            elif self.cropping[1][1] == self.cropping[2][1] == 0:
                return inputs[
                    :,
                    :,
                    self.cropping[0][0] : -self.cropping[0][1],
                    self.cropping[1][0] :,
                    self.cropping[2][0] :,
                ]
            elif self.cropping[0][1] == self.cropping[2][1] == 0:
                return inputs[
                    :,
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] : -self.cropping[1][1],
                    self.cropping[2][0] :,
                ]
            elif self.cropping[0][1] == 0:
                return inputs[
                    :,
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] : -self.cropping[1][1],
                    self.cropping[2][0] : -self.cropping[2][1],
                ]
            elif self.cropping[1][1] == 0:
                return inputs[
                    :,
                    :,
                    self.cropping[0][0] : -self.cropping[0][1],
                    self.cropping[1][0] :,
                    self.cropping[2][0] : -self.cropping[2][1],
                ]
            elif self.cropping[2][1] == 0:
                return inputs[
                    :,
                    :,
                    self.cropping[0][0] : -self.cropping[0][1],
                    self.cropping[1][0] : -self.cropping[1][1],
                    self.cropping[2][0] :,
                ]
            return inputs[
                :,
                :,
                self.cropping[0][0] : -self.cropping[0][1],
                self.cropping[1][0] : -self.cropping[1][1],
                self.cropping[2][0] : -self.cropping[2][1],
            ]
        else:
            if (
                self.cropping[0][1]
                == self.cropping[1][1]
                == self.cropping[2][1]
                == 0
            ):
                return inputs[
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] :,
                    self.cropping[2][0] :,
                    :,
                ]
            elif self.cropping[0][1] == self.cropping[1][1] == 0:
                return inputs[
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] :,
                    self.cropping[2][0] : -self.cropping[2][1],
                    :,
                ]
            elif self.cropping[1][1] == self.cropping[2][1] == 0:
                return inputs[
                    :,
                    self.cropping[0][0] : -self.cropping[0][1],
                    self.cropping[1][0] :,
                    self.cropping[2][0] :,
                    :,
                ]
            elif self.cropping[0][1] == self.cropping[2][1] == 0:
                return inputs[
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] : -self.cropping[1][1],
                    self.cropping[2][0] :,
                    :,
                ]
            elif self.cropping[0][1] == 0:
                return inputs[
                    :,
                    self.cropping[0][0] :,
                    self.cropping[1][0] : -self.cropping[1][1],
                    self.cropping[2][0] : -self.cropping[2][1],
                    :,
                ]
            elif self.cropping[1][1] == 0:
                return inputs[
                    :,
                    self.cropping[0][0] : -self.cropping[0][1],
                    self.cropping[1][0] :,
                    self.cropping[2][0] : -self.cropping[2][1],
                    :,
                ]
            elif self.cropping[2][1] == 0:
                return inputs[
                    :,
                    self.cropping[0][0] : -self.cropping[0][1],
                    self.cropping[1][0] : -self.cropping[1][1],
                    self.cropping[2][0] :,
                    :,
                ]
            return inputs[
                :,
                self.cropping[0][0] : -self.cropping[0][1],
                self.cropping[1][0] : -self.cropping[1][1],
                self.cropping[2][0] : -self.cropping[2][1],
                :,
            ]

    def get_config(self):
        config = {"cropping": self.cropping, "data_format": self.data_format}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Built-in optimizer classes.

For more examples see the base class `tf.keras.optimizers.Optimizer`.
"""

# Imports needed for deserialization.

import platform
import warnings

import tensorflow.compat.v2 as tf
from absl import logging

from keras.src import backend
from keras.src.optimizers import adadelta
from keras.src.optimizers import adafactor
from keras.src.optimizers import adagrad
from keras.src.optimizers import adam
from keras.src.optimizers import adamax
from keras.src.optimizers import adamw
from keras.src.optimizers import ftrl
from keras.src.optimizers import lion
from keras.src.optimizers import nadam
from keras.src.optimizers import optimizer as base_optimizer
from keras.src.optimizers import rmsprop
from keras.src.optimizers import sgd
from keras.src.optimizers.legacy import adadelta as adadelta_legacy
from keras.src.optimizers.legacy import adagrad as adagrad_legacy
from keras.src.optimizers.legacy import adam as adam_legacy
from keras.src.optimizers.legacy import adamax as adamax_legacy
from keras.src.optimizers.legacy import ftrl as ftrl_legacy
from keras.src.optimizers.legacy import gradient_descent as gradient_descent_legacy
from keras.src.optimizers.legacy import nadam as nadam_legacy
from keras.src.optimizers.legacy import optimizer_v2 as base_optimizer_legacy
from keras.src.optimizers.legacy import rmsprop as rmsprop_legacy
from keras.src.optimizers.legacy.adadelta import Adadelta
from keras.src.optimizers.legacy.adagrad import Adagrad
from keras.src.optimizers.legacy.adam import Adam
from keras.src.optimizers.legacy.adamax import Adamax
from keras.src.optimizers.legacy.ftrl import Ftrl

# Symbols to be accessed under keras.optimizers. To be replaced with
# optimizers v2022 when they graduate out of experimental.
from keras.src.optimizers.legacy.gradient_descent import SGD
from keras.src.optimizers.legacy.nadam import Nadam
from keras.src.optimizers.legacy.rmsprop import RMSprop
from keras.src.optimizers.optimizer_v1 import Optimizer
from keras.src.optimizers.optimizer_v1 import TFOptimizer
from keras.src.optimizers.schedules import learning_rate_schedule
from keras.src.saving.legacy import serialization as legacy_serialization
from keras.src.saving.serialization_lib import deserialize_keras_object
from keras.src.saving.serialization_lib import serialize_keras_object

# isort: off
from tensorflow.python.util.tf_export import keras_export

# pylint: disable=line-too-long


@keras_export("keras.optimizers.serialize")
def serialize(optimizer, use_legacy_format=False):
    """Serialize the optimizer configuration to JSON compatible python dict.

    The configuration can be used for persistence and reconstruct the
    `Optimizer` instance again.

    >>> tf.keras.optimizers.serialize(tf.keras.optimizers.legacy.SGD())
    {'module': 'keras.optimizers.legacy', 'class_name': 'SGD', 'config': {'name': 'SGD', 'learning_rate': 0.01, 'decay': 0.0, 'momentum': 0.0, 'nesterov': False}, 'registered_name': None}"""  # noqa: E501
    """
    Args:
      optimizer: An `Optimizer` instance to serialize.

    Returns:
      Python dict which contains the configuration of the input optimizer.
    """
    if optimizer is None:
        return None
    if not isinstance(
        optimizer,
        (
            base_optimizer.Optimizer,
            Optimizer,
            base_optimizer_legacy.OptimizerV2,
        ),
    ):
        warnings.warn(
            "The `keras.optimizers.serialize()` API should only be used for "
            "objects of type `keras.optimizers.Optimizer`. Found an instance "
            f"of type {type(optimizer)}, which may lead to improper "
            "serialization."
        )
    if use_legacy_format:
        return legacy_serialization.serialize_keras_object(optimizer)
    return serialize_keras_object(optimizer)


def is_arm_mac():
    return platform.system() == "Darwin" and platform.processor() == "arm"


@keras_export("keras.optimizers.deserialize")
def deserialize(config, custom_objects=None, use_legacy_format=False, **kwargs):
    """Inverse of the `serialize` function.

    Args:
        config: Optimizer configuration dictionary.
        custom_objects: Optional dictionary mapping names (strings) to custom
            objects (classes and functions) to be considered during
            deserialization.

    Returns:
        A Keras Optimizer instance.
    """
    # loss_scale_optimizer has a direct dependency of optimizer, import here
    # rather than top to avoid the cyclic dependency.
    from keras.src.mixed_precision import (
        loss_scale_optimizer,
    )

    use_legacy_optimizer = kwargs.pop("use_legacy_optimizer", False)
    if kwargs:
        raise TypeError(f"Invalid keyword arguments: {kwargs}")
    if len(config["config"]) > 0:
        # If the optimizer config is not empty, then we use the value of
        # `is_legacy_optimizer` to override `use_legacy_optimizer`. If
        # `is_legacy_optimizer` does not exist in config, it means we are
        # using the legacy optimzier.
        use_legacy_optimizer = config["config"].get("is_legacy_optimizer", True)
    if (
        tf.__internal__.tf2.enabled()
        and tf.executing_eagerly()
        and not is_arm_mac()
        and not use_legacy_optimizer
    ):
        # We observed a slowdown of optimizer on M1 Mac, so we fall back to the
        # legacy optimizer for M1 users now, see b/263339144 for more context.
        all_classes = {
            "adadelta": adadelta.Adadelta,
            "adagrad": adagrad.Adagrad,
            "adam": adam.Adam,
            "adamax": adamax.Adamax,
            "experimentaladadelta": adadelta.Adadelta,
            "experimentaladagrad": adagrad.Adagrad,
            "experimentaladam": adam.Adam,
            "experimentalsgd": sgd.SGD,
            "nadam": nadam.Nadam,
            "rmsprop": rmsprop.RMSprop,
            "sgd": sgd.SGD,
            "ftrl": ftrl.Ftrl,
            "lossscaleoptimizer": loss_scale_optimizer.LossScaleOptimizerV3,
            "lossscaleoptimizerv3": loss_scale_optimizer.LossScaleOptimizerV3,
            # LossScaleOptimizerV1 was an old version of LSO that was removed.
            # Deserializing it turns it into a LossScaleOptimizer
            "lossscaleoptimizerv1": loss_scale_optimizer.LossScaleOptimizer,
        }
    else:
        all_classes = {
            "adadelta": adadelta_legacy.Adadelta,
            "adagrad": adagrad_legacy.Adagrad,
            "adam": adam_legacy.Adam,
            "adamax": adamax_legacy.Adamax,
            "experimentaladadelta": adadelta.Adadelta,
            "experimentaladagrad": adagrad.Adagrad,
            "experimentaladam": adam.Adam,
            "experimentalsgd": sgd.SGD,
            "nadam": nadam_legacy.Nadam,
            "rmsprop": rmsprop_legacy.RMSprop,
            "sgd": gradient_descent_legacy.SGD,
            "ftrl": ftrl_legacy.Ftrl,
            "lossscaleoptimizer": loss_scale_optimizer.LossScaleOptimizer,
            "lossscaleoptimizerv3": loss_scale_optimizer.LossScaleOptimizerV3,
            # LossScaleOptimizerV1 was an old version of LSO that was removed.
            # Deserializing it turns it into a LossScaleOptimizer
            "lossscaleoptimizerv1": loss_scale_optimizer.LossScaleOptimizer,
        }

    # Make deserialization case-insensitive for built-in optimizers.
    if config["class_name"].lower() in all_classes:
        config["class_name"] = config["class_name"].lower()

    if use_legacy_format:
        return legacy_serialization.deserialize_keras_object(
            config,
            module_objects=all_classes,
            custom_objects=custom_objects,
            printable_module_name="optimizer",
        )

    return deserialize_keras_object(
        config,
        module_objects=all_classes,
        custom_objects=custom_objects,
        printable_module_name="optimizer",
    )


@keras_export(
    "keras.__internal__.optimizers.convert_to_legacy_optimizer", v1=[]
)
def convert_to_legacy_optimizer(optimizer):
    """Convert experimental optimizer to legacy optimizer.

    This function takes in a `keras.optimizers.Optimizer`
    instance and converts it to the corresponding
    `keras.optimizers.legacy.Optimizer` instance.
    For example, `keras.optimizers.Adam(...)` to
    `keras.optimizers.legacy.Adam(...)`.

    Args:
        optimizer: An instance of `keras.optimizers.Optimizer`.
    """
    # loss_scale_optimizer has a direct dependency of optimizer, import here
    # rather than top to avoid the cyclic dependency.
    from keras.src.mixed_precision import (
        loss_scale_optimizer,
    )

    if not isinstance(optimizer, base_optimizer.Optimizer):
        raise ValueError(
            "`convert_to_legacy_optimizer` should only be called "
            "on instances of `tf.keras.optimizers.Optimizer`, but "
            f"received {optimizer} of type {type(optimizer)}."
        )
    optimizer_name = optimizer.__class__.__name__.lower()
    config = optimizer.get_config()
    # Remove fields that only exist in experimental optimizer.
    keys_to_remove = [
        "weight_decay",
        "use_ema",
        "ema_momentum",
        "ema_overwrite_frequency",
        "jit_compile",
        "is_legacy_optimizer",
    ]
    for key in keys_to_remove:
        config.pop(key, None)

    if isinstance(optimizer, loss_scale_optimizer.LossScaleOptimizerV3):
        # For LossScaleOptimizers, recursively convert the inner optimizer
        config["inner_optimizer"] = convert_to_legacy_optimizer(
            optimizer.inner_optimizer
        )
        if optimizer_name == "lossscaleoptimizerv3":
            optimizer_name = "lossscaleoptimizer"

    # Learning rate can be a custom LearningRateSchedule, which is stored as
    # a dict in config, and cannot be deserialized.
    if hasattr(optimizer, "_learning_rate") and isinstance(
        optimizer._learning_rate, learning_rate_schedule.LearningRateSchedule
    ):
        config["learning_rate"] = optimizer._learning_rate
    legacy_optimizer_config = {
        "class_name": optimizer_name,
        "config": config,
    }
    return deserialize(legacy_optimizer_config, use_legacy_optimizer=True)


@keras_export("keras.optimizers.get")
def get(identifier, **kwargs):
    """Retrieves a Keras Optimizer instance.

    Args:
        identifier: Optimizer identifier, one of - String: name of an optimizer
          - Dictionary: configuration dictionary. - Keras Optimizer instance (it
          will be returned unchanged). - TensorFlow Optimizer instance (it will
          be wrapped as a Keras Optimizer).

    Returns:
        A Keras Optimizer instance.

    Raises:
        ValueError: If `identifier` cannot be interpreted.
    """
    use_legacy_optimizer = kwargs.pop("use_legacy_optimizer", False)
    if kwargs:
        raise TypeError(f"Invalid keyword arguments: {kwargs}")
    if isinstance(
        identifier,
        (
            Optimizer,
            base_optimizer_legacy.OptimizerV2,
        ),
    ):
        return identifier
    elif isinstance(identifier, base_optimizer.Optimizer):
        if tf.__internal__.tf2.enabled() and not is_arm_mac():
            return identifier
        else:
            # If TF2 is disabled or on a M1 mac, we convert to the legacy
            # optimizer. We observed a slowdown of optimizer on M1 Mac, so we
            # fall back to the legacy optimizer for now, see b/263339144
            # for more context.
            optimizer_name = identifier.__class__.__name__
            logging.warning(
                "There is a known slowdown when using v2.11+ Keras optimizers "
                "on M1/M2 Macs. Falling back to the "
                "legacy Keras optimizer, i.e., "
                f"`tf.keras.optimizers.legacy.{optimizer_name}`."
            )
            return convert_to_legacy_optimizer(identifier)

    # Wrap legacy TF optimizer instances
    elif isinstance(identifier, tf.compat.v1.train.Optimizer):
        opt = TFOptimizer(identifier)
        backend.track_tf_optimizer(opt)
        return opt
    elif isinstance(identifier, dict):
        use_legacy_format = "module" not in identifier
        return deserialize(
            identifier,
            use_legacy_optimizer=use_legacy_optimizer,
            use_legacy_format=use_legacy_format,
        )
    elif isinstance(identifier, str):
        config = {"class_name": str(identifier), "config": {}}
        return get(
            config,
            use_legacy_optimizer=use_legacy_optimizer,
        )
    else:
        raise ValueError(
            f"Could not interpret optimizer identifier: {identifier}"
        )




"""A binary that creates a serialized SavedModel from a keras model.

This is used in tests to ensure that model serialization is deterministic across
different processes.
"""

import tensorflow.compat.v2 as tf
from absl import app
from absl import flags

from keras.src import regularizers
from keras.src.testing_infra import test_utils

flags.DEFINE_string("output_path", "", "The path to write the SavedModel at.")

FLAGS = flags.FLAGS


def main(_) -> None:
    with test_utils.model_type_scope("functional"):
        model = test_utils.get_small_mlp(1, 4, input_dim=3)
        model.layers[-1].activity_regularizer = regularizers.get("l2")
        model.activity_regularizer = regularizers.get("l2")
        model.compile(loss="mse", optimizer="rmsprop")

        def callable_loss():
            return tf.reduce_sum(model.weights[0])

        model.add_loss(callable_loss)

        print(f"_____Writing saved model to: {FLAGS.output_path}")
        model.save(FLAGS.output_path)


if __name__ == "__main__":
    app.run(main)




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for preprocessing sequence data.

Deprecated: `tf.keras.preprocessing.sequence` APIs are not recommended for new
code. Prefer `tf.keras.utils.timeseries_dataset_from_array` and
the `tf.data` APIs which provide a much more flexible mechanisms for dealing
with sequences. See the [tf.data guide](https://www.tensorflow.org/guide/data)
for more details.
"""


import json
import random

import numpy as np

from keras.src.utils import data_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


def _remove_long_seq(maxlen, seq, label):
    """Removes sequences that exceed the maximum length.

    Args:
        maxlen: Int, maximum length of the output sequences.
        seq: List of lists, where each sublist is a sequence.
        label: List where each element is an integer.

    Returns:
        new_seq, new_label: shortened lists for `seq` and `label`.
    """
    new_seq, new_label = [], []
    for x, y in zip(seq, label):
        if len(x) < maxlen:
            new_seq.append(x)
            new_label.append(y)
    return new_seq, new_label


@keras_export("keras.preprocessing.sequence.TimeseriesGenerator")
class TimeseriesGenerator(data_utils.Sequence):
    """Utility class for generating batches of temporal data.

    Deprecated: `tf.keras.preprocessing.sequence.TimeseriesGenerator` does not
    operate on tensors and is not recommended for new code. Prefer using a
    `tf.data.Dataset` which provides a more efficient and flexible mechanism for
    batching, shuffling, and windowing input. See the
    [tf.data guide](https://www.tensorflow.org/guide/data) for more details.

    This class takes in a sequence of data-points gathered at
    equal intervals, along with time series parameters such as
    stride, length of history, etc., to produce batches for
    training/validation.

    Arguments:
        data: Indexable generator (such as list or Numpy array)
            containing consecutive data points (timesteps).
            The data should be at 2D, and axis 0 is expected
            to be the time dimension.
        targets: Targets corresponding to timesteps in `data`.
            It should have same length as `data`.
        length: Length of the output sequences (in number of timesteps).
        sampling_rate: Period between successive individual timesteps
            within sequences. For rate `r`, timesteps
            `data[i]`, `data[i-r]`, ... `data[i - length]`
            are used for create a sample sequence.
        stride: Period between successive output sequences.
            For stride `s`, consecutive output samples would
            be centered around `data[i]`, `data[i+s]`, `data[i+2*s]`, etc.
        start_index: Data points earlier than `start_index` will not be used
            in the output sequences. This is useful to reserve part of the
            data for test or validation.
        end_index: Data points later than `end_index` will not be used
            in the output sequences. This is useful to reserve part of the
            data for test or validation.
        shuffle: Whether to shuffle output samples,
            or instead draw them in chronological order.
        reverse: Boolean: if `true`, timesteps in each output sample will be
            in reverse chronological order.
        batch_size: Number of timeseries samples in each batch
            (except maybe the last one).

    Returns:
        A [Sequence](
        https://www.tensorflow.org/api_docs/python/tf/keras/utils/Sequence)
        instance.

    Examples:
        ```python
        from keras.src.preprocessing.sequence import TimeseriesGenerator
        import numpy as np
        data = np.array([[i] for i in range(50)])
        targets = np.array([[i] for i in range(50)])
        data_gen = TimeseriesGenerator(data, targets,
                                       length=10, sampling_rate=2,
                                       batch_size=2)
        assert len(data_gen) == 20
        batch_0 = data_gen[0]
        x, y = batch_0
        assert np.array_equal(x,
                              np.array([[[0], [2], [4], [6], [8]],
                                        [[1], [3], [5], [7], [9]]]))
        assert np.array_equal(y,
                              np.array([[10], [11]]))
        ```
    """

    def __init__(
        self,
        data,
        targets,
        length,
        sampling_rate=1,
        stride=1,
        start_index=0,
        end_index=None,
        shuffle=False,
        reverse=False,
        batch_size=128,
    ):

        if len(data) != len(targets):
            raise ValueError(
                "Data and targets have to be"
                + f" of same length. Data length is {len(data)}"
                + f" while target length is {len(targets)}"
            )

        self.data = data
        self.targets = targets
        self.length = length
        self.sampling_rate = sampling_rate
        self.stride = stride
        self.start_index = start_index + length
        if end_index is None:
            end_index = len(data) - 1
        self.end_index = end_index
        self.shuffle = shuffle
        self.reverse = reverse
        self.batch_size = batch_size

        if self.start_index > self.end_index:
            raise ValueError(
                "`start_index+length=%i > end_index=%i` "
                "is disallowed, as no part of the sequence "
                "would be left to be used as current step."
                % (self.start_index, self.end_index)
            )

    def __len__(self):
        return (
            self.end_index - self.start_index + self.batch_size * self.stride
        ) // (self.batch_size * self.stride)

    def __getitem__(self, index):
        if self.shuffle:
            rows = np.random.randint(
                self.start_index, self.end_index + 1, size=self.batch_size
            )
        else:
            i = self.start_index + self.batch_size * self.stride * index
            rows = np.arange(
                i,
                min(i + self.batch_size * self.stride, self.end_index + 1),
                self.stride,
            )

        samples = np.array(
            [
                self.data[row - self.length : row : self.sampling_rate]
                for row in rows
            ]
        )
        targets = np.array([self.targets[row] for row in rows])

        if self.reverse:
            return samples[:, ::-1, ...], targets
        return samples, targets

    def get_config(self):
        """Returns the TimeseriesGenerator configuration as Python dictionary.

        Returns:
            A Python dictionary with the TimeseriesGenerator configuration.
        """
        data = self.data
        if type(self.data).__module__ == np.__name__:
            data = self.data.tolist()
        try:
            json_data = json.dumps(data)
        except TypeError as e:
            raise TypeError("Data not JSON Serializable:", data) from e

        targets = self.targets
        if type(self.targets).__module__ == np.__name__:
            targets = self.targets.tolist()
        try:
            json_targets = json.dumps(targets)
        except TypeError as e:
            raise TypeError("Targets not JSON Serializable:", targets) from e

        return {
            "data": json_data,
            "targets": json_targets,
            "length": self.length,
            "sampling_rate": self.sampling_rate,
            "stride": self.stride,
            "start_index": self.start_index,
            "end_index": self.end_index,
            "shuffle": self.shuffle,
            "reverse": self.reverse,
            "batch_size": self.batch_size,
        }

    def to_json(self, **kwargs):
        """Returns a JSON string containing the generator's configuration.

        Args:
            **kwargs: Additional keyword arguments to be passed
                to `json.dumps()`.

        Returns:
            A JSON string containing the tokenizer configuration.
        """
        config = self.get_config()
        timeseries_generator_config = {
            "class_name": self.__class__.__name__,
            "config": config,
        }
        return json.dumps(timeseries_generator_config, **kwargs)


@keras_export("keras.preprocessing.sequence.make_sampling_table")
def make_sampling_table(size, sampling_factor=1e-5):
    """Generates a word rank-based probabilistic sampling table.

    Used for generating the `sampling_table` argument for `skipgrams`.
    `sampling_table[i]` is the probability of sampling
    the word i-th most common word in a dataset
    (more common words should be sampled less frequently, for balance).

    The sampling probabilities are generated according
    to the sampling distribution used in word2vec:

    ```
    p(word) = (min(1, sqrt(word_frequency / sampling_factor) /
        (word_frequency / sampling_factor)))
    ```

    We assume that the word frequencies follow Zipf's law (s=1) to derive
    a numerical approximation of frequency(rank):

    `frequency(rank) ~ 1/(rank * (log(rank) + gamma) + 1/2 - 1/(12*rank))`
    where `gamma` is the Euler-Mascheroni constant.

    Args:
        size: Int, number of possible words to sample.
        sampling_factor: The sampling factor in the word2vec formula.

    Returns:
        A 1D Numpy array of length `size` where the ith entry
        is the probability that a word of rank i should be sampled.
    """
    gamma = 0.577
    rank = np.arange(size)
    rank[0] = 1
    inv_fq = rank * (np.log(rank) + gamma) + 0.5 - 1.0 / (12.0 * rank)
    f = sampling_factor * inv_fq

    return np.minimum(1.0, f / np.sqrt(f))


@keras_export("keras.preprocessing.sequence.skipgrams")
def skipgrams(
    sequence,
    vocabulary_size,
    window_size=4,
    negative_samples=1.0,
    shuffle=True,
    categorical=False,
    sampling_table=None,
    seed=None,
):
    """Generates skipgram word pairs.

    This function transforms a sequence of word indexes (list of integers)
    into tuples of words of the form:

    - (word, word in the same window), with label 1 (positive samples).
    - (word, random word from the vocabulary), with label 0 (negative samples).

    Read more about Skipgram in this gnomic paper by Mikolov et al.:
    [Efficient Estimation of Word Representations in
    Vector Space](http://arxiv.org/pdf/1301.3781v3.pdf)

    Args:
        sequence: A word sequence (sentence), encoded as a list
            of word indices (integers). If using a `sampling_table`,
            word indices are expected to match the rank
            of the words in a reference dataset (e.g. 10 would encode
            the 10-th most frequently occurring token).
            Note that index 0 is expected to be a non-word and will be skipped.
        vocabulary_size: Int, maximum possible word index + 1
        window_size: Int, size of sampling windows (technically half-window).
            The window of a word `w_i` will be
            `[i - window_size, i + window_size+1]`.
        negative_samples: Float >= 0. 0 for no negative (i.e. random) samples.
            1 for same number as positive samples.
        shuffle: Whether to shuffle the word couples before returning them.
        categorical: bool. if False, labels will be
            integers (eg. `[0, 1, 1 .. ]`),
            if `True`, labels will be categorical, e.g.
            `[[1,0],[0,1],[0,1] .. ]`.
        sampling_table: 1D array of size `vocabulary_size` where the entry i
            encodes the probability to sample a word of rank i.
        seed: Random seed.

    Returns:
        couples, labels: where `couples` are int pairs and
            `labels` are either 0 or 1.

    Note:
        By convention, index 0 in the vocabulary is
        a non-word and will be skipped.
    """
    couples = []
    labels = []
    for i, wi in enumerate(sequence):
        if not wi:
            continue
        if sampling_table is not None:
            if sampling_table[wi] < random.random():
                continue

        window_start = max(0, i - window_size)
        window_end = min(len(sequence), i + window_size + 1)
        for j in range(window_start, window_end):
            if j != i:
                wj = sequence[j]
                if not wj:
                    continue
                couples.append([wi, wj])
                if categorical:
                    labels.append([0, 1])
                else:
                    labels.append(1)

    if negative_samples > 0:
        num_negative_samples = int(len(labels) * negative_samples)
        words = [c[0] for c in couples]
        random.shuffle(words)

        couples += [
            [words[i % len(words)], random.randint(1, vocabulary_size - 1)]
            for i in range(num_negative_samples)
        ]
        if categorical:
            labels += [[1, 0]] * num_negative_samples
        else:
            labels += [0] * num_negative_samples

    if shuffle:
        if seed is None:
            seed = random.randint(0, 10e6)
        random.seed(seed)
        random.shuffle(couples)
        random.seed(seed)
        random.shuffle(labels)

    return couples, labels




"""AUTOGENERATED. DO NOT EDIT."""

from keras.src.applications.mobilenet import MobileNet
from keras.src.applications.mobilenet import decode_predictions
from keras.src.applications.mobilenet import preprocess_input
